from django.apps import AppConfig, apps
from .registries import (
    user_menu_registry,
    navigation
)


class AdminConfig(AppConfig):
    name = 'papermerge.contrib.admin'
    label = 'admin'

    def ready(self):
        from papermerge.contrib.admin import signals  # noqa

        # autodiscover user menu items from user_menu.py modules
        app_names = [app.name for app in apps.app_configs.values()]
        user_menu_registry.autodiscover(app_names)
        navigation.autodiscover(app_names)



# Generated by Django 3.0.2 on 2020-02-25 07:18

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0003_document_version'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='document',
            name='celery_task_id',
        ),
    ]



import json
import logging

from django.http import (
    HttpResponseBadRequest,
    HttpResponseForbidden,
    Http404,
    HttpResponse
)
from django.contrib.auth.decorators import login_required
from django.views.decorators.http import require_GET
from django.urls import reverse
from django.shortcuts import (
    get_object_or_404,
    redirect
)
from django.utils.translation import gettext as _
from django.core.files.temp import NamedTemporaryFile
from django.core.paginator import Paginator
from django.db.models.functions import Lower

from mglib import mime

from papermerge.core.models import (
    BaseTreeNode,
    Access,
    Folder,
    Automate,
    Tag
)
from papermerge.core import signal_definitions as signals
from papermerge.core.backup_restore import build_tar_archive
from papermerge.core.storage import default_storage

from .decorators import json_response
from .utils import sanitize_kvstore_list

logger = logging.getLogger(__name__)


PER_PAGE = 30


def _filter_by_tag(nodes, request_get_dict):
    tag = request_get_dict.get('tag', None)

    if tag:
        nodes = BaseTreeNode.objects.filter(
            tags__name__in=[tag]
        ).exclude(
            title=Folder.INBOX_NAME
        )

    return nodes


def _order_by(nodes, request_get_dict):

    order_by = request_get_dict.get('order-by', '-type')

    # order_by might be empty string
    if not order_by:
        # instead of empty string, use default
        # order by type descending
        order_by = '-type'

    # Order by these fields. Presence of dash (minus character)
    # in front of the field toggles ascending or descending order (django way).
    if order_by in ('date', '-date', 'title', '-title', 'type', '-type'):
        field_map = {
            'type': 'polymorphic_ctype_id',
            '-type': '-polymorphic_ctype_id',
            'date': 'created_at',
            '-date': '-created_at',
            'title': 'title',
            '-title': '-title'
        }
        field = field_map[order_by]
        # case consistent ordering of title field
        if field == 'title':
            nodes = nodes.order_by(Lower('title').asc())
        elif field == '-title':
            nodes = nodes.order_by(Lower('title').desc())
        else:
            nodes = nodes.order_by(field)

    return nodes


def _get_node_kv(node_id) -> list:
    kv_list = []

    if node_id:
        parent_node = get_object_or_404(
            BaseTreeNode, id=node_id
        )

        for item in parent_node.kv.all():
            kv_list.append(item.to_dict())

    return kv_list


def _get_page_number(request_get) -> int:
    page_number = request_get.get('page', 1)

    if not page_number:
        page_number = 1

    return int(page_number)


def _get_pagination_dict(paginator, page_number) -> dict:

    page_obj = paginator.get_page(page_number)

    num_pages = paginator.num_pages

    # 1.   Number of pages < 7: show all pages;
    # 2.   Current page <= 4: show first 7 pages;
    # 3.   Current page > 4 and < (number of pages - 4): show current page,
    #       3 before and 3 after;
    # 4.   Current page >= (number of pages - 4): show the last 7 pages.

    if num_pages <= 7 or page_number <= 4:  # case 1 and 2
        pages = [x for x in range(1, min(num_pages + 1, 7))]
    elif page_number > num_pages - 4:  # case 4
        pages = [x for x in range(num_pages - 6, num_pages + 1)]
    else:  # case 3
        pages = [x for x in range(page_number - 3, page_number + 4)]

    if page_obj.has_previous():
        previous_page_number = page_obj.previous_page_number()
    else:
        previous_page_number = -1

    if page_obj.has_next():
        next_page_number = page_obj.next_page_number()
    else:
        next_page_number = -1

    return {
        'page_number': page_number,
        'pages': pages,
        'num_pages': num_pages,
        'page': {
            'has_previous': page_obj.has_previous(),
            'has_next': page_obj.has_next(),
            'previous_page_number': previous_page_number,
            'next_page_number': next_page_number,
        }
    }


@json_response
@login_required
@require_GET
def browse_view(request, parent_id=None):
    """
    Returns json string of nodes which current user can read:

        * filtered by tag (optionally)
        * paginated
        * ordered by title, date, type (optionally)
    """
    nodes = BaseTreeNode.objects.filter(parent_id=parent_id).exclude(
        title=Folder.INBOX_NAME
    )

    nodes = _filter_by_tag(nodes, request.GET)
    nodes = _order_by(nodes, request.GET)

    nodes_list = []
    nodes_perms = request.user.get_perms_dict(
        nodes, Access.ALL_PERMS
    )
    readable_nodes = []
    for node in nodes:
        if nodes_perms[node.id].get(Access.PERM_READ, False):
            readable_nodes.append(node)

    page_number = _get_page_number(request_get=request.GET)
    paginator = Paginator(readable_nodes, PER_PAGE)
    page_obj = paginator.get_page(page_number)

    for node in page_obj.object_list:
        node_dict = node.to_dict()
        # and send user_perms to the frontend client

        node_dict['user_perms'] = nodes_perms[node.id]

        if node.is_document():
            node_dict['img_src'] = reverse(
                'core:preview',
                args=(node.id, 4, 1)
            )
            node_dict['document_url'] = reverse(
                'core:document',
                args=(node.id,)
            )

        nodes_list.append(node_dict)

    return {
        'nodes': nodes_list,
        'parent_id': parent_id,
        'parent_kv': _get_node_kv(parent_id),
        'pagination': _get_pagination_dict(
            paginator=paginator,
            page_number=page_number
        )
    }


@json_response
@login_required
def breadcrumb_view(request, parent_id=None):

    nodes = []

    node = None
    try:
        node = BaseTreeNode.objects.get(id=parent_id)
    except BaseTreeNode.DoesNotExist:
        pass

    if node:
        nodes = [
            item.to_dict() for item in node.get_ancestors(include_self=True)
        ]

    return {
        'nodes': nodes,
    }


@json_response
@login_required
def node_by_title_view(request, title):
    """
    Useful in case of special folders like inbox (and trash in future).
    Returns node id, children_count, title of specified node's title.
    Title specified is insensitive (i.e. INBOX = Inbox = inbox).
    """
    node = get_object_or_404(
        BaseTreeNode,
        title__iexact=title
    )

    return {
        'id': node.id,
        'title': node.title,
        'children_count': node.get_children().count(),
        'url': reverse('node_by_title', args=('inbox',))
    }


@json_response
@login_required
def node_view(request, node_id):
    """
    GET or DELETE /node/<node_id>
    """
    try:
        node = BaseTreeNode.objects.get(id=node_id)
    except BaseTreeNode.DoesNotExist:
        ret = {
            'node': node.to_dict()
        }
        return ret, HttpResponseBadRequest.status_code

    if request.method == "DELETE":
        if request.user.has_perm(Access.PERM_DELETE, node):
            node.delete()
        else:
            msg = f"{request.user.username} does not have" +\
                f" permission to delete {node.title}"

            return msg, HttpResponseForbidden.status_code

        return 'OK'

    if request.method in ("PUT", "POST"):
        if not request.user.has_perm(Access.PERM_WRITE, node):
            return "Permission denied", HttpResponseForbidden.status_code

        node_data = json.loads(request.body)
        if 'kvstore' in node_data:
            metadata = node_data['kvstore']
            if isinstance(metadata, list):
                node.kv.update(
                    sanitize_kvstore_list(metadata)
                )

    if not request.user.has_perm(Access.PERM_READ, node):
        return "Permission denied", HttpResponseForbidden.status_code

    node_dict = node.to_dict()
    # minor hack to enable autocomplete for tag editor
    # in document view
    node_dict['alltags'] = [
        tag.to_dict()
        for tag in Tag.objects.filter(user=request.user)
    ]

    return {
        'node': node_dict
    }


@json_response
@login_required
def nodes_view(request):
    """
    GET or POST /nodes/
    """

    if request.method == "POST":

        data = json.loads(request.body)
        node_ids = [item['id'] for item in data]
        queryset = BaseTreeNode.objects.filter(id__in=node_ids)

        automates = Automate.objects.filter(
            dst_folder__in=queryset
        )
        if automates.count():
            msg = _(
                "Following Automates have references to folders "
                "you are trying to delete: "
            )
            msg += ", ".join([auto.name for auto in automates])
            msg += _(
                ". Please delete mentioned Automates first."
            )

            return msg, HttpResponseBadRequest.status_code

        nodes_perms = request.user.get_perms_dict(
            queryset, Access.ALL_PERMS
        )
        node_titles = []
        for node in queryset:
            node_titles.append(
                node.title
            )
            if not nodes_perms[node.id].get(Access.PERM_DELETE, False):
                # if user does not have delete permission on
                # one node - forbid entire operation!
                msg = f"{request.user.username} does not have" +\
                    f" permission to delete {node.title}"

                return msg, HttpResponseForbidden.status_code
        # yes, user is allowed to delete all nodes,
        # proceed with delete opration
        queryset.delete()
        signals.nodes_deleted.send(
            sender='core.views.nodes.nodes_view',
            user_id=request.user.id,
            level=logging.INFO,
            message=_("Nodes deleted"),
            node_titles=node_titles,
            node_ids=node_ids
        )

        return 'OK'

    return 'OK'


@login_required
def node_download(request, id):
    """
    Any user with read permission on the node must be
    able to download it.

    Node is either documennt or a folder.
    """
    version = request.GET.get('version', None)

    try:
        node = BaseTreeNode.objects.get(id=id)
    except BaseTreeNode.DoesNotExist:
        raise Http404("Node does not exists")

    if request.user.has_perm(Access.PERM_READ, node):

        if node.is_document():
            file_abs_path = default_storage.abspath(
                node.path().url(version=version)
            )
            mime_type = mime.Mime(file_abs_path)
            try:
                file_handle = open(file_abs_path, "rb")
            except OSError:
                logger.error(
                    "Cannot open local version of %s" % node.path.url()
                )
                return redirect('admin:browse')

            resp = HttpResponse(
                file_handle.read(),
                content_type=mime_type.guess()
            )
            disposition = "attachment; filename=%s" % node.title
            resp['Content-Disposition'] = disposition
            file_handle.close()

            return resp
        else:  # node is a folder

            with NamedTemporaryFile(prefix="download_") as fileobj:
                # collected into an archive all direct children of
                # selected folder
                node_ids = [_node.id for _node in node.get_children()]
                build_tar_archive(
                    fileobj=fileobj,
                    node_ids=node_ids
                )
                # reset fileobj to initial position
                fileobj.seek(0)
                data = fileobj.read()
                resp = HttpResponse(
                    data,
                    content_type="application/x-tar"
                )
                disposition = f"attachment; filename={node.title}.tar"
                resp['Content-Disposition'] = disposition

                return resp

    return HttpResponseForbidden()


@login_required
def nodes_download(request):
    """
    Download multiple nodes (documents and folders) packed
    as tar.gz archive.
    """

    node_ids = request.GET.getlist('node_ids[]')
    nodes = BaseTreeNode.objects.filter(
        id__in=node_ids
    )
    nodes_perms = request.user.get_perms_dict(
        nodes, Access.ALL_PERMS
    )
    for node in nodes:
        if not nodes_perms[node.id].get(
            Access.PERM_READ, False
        ):
            msg = _(
                "%s does not have permission to read %s"
            ) % (request.user.username, node.title)

            return msg, HttpResponseForbidden.status_code

    with NamedTemporaryFile(prefix="download_") as fileobj:
        build_tar_archive(
            fileobj=fileobj,
            node_ids=node_ids
        )
        # reset fileobj to initial position
        fileobj.seek(0)
        data = fileobj.read()
        resp = HttpResponse(
            data,
            content_type="application/x-tar"
        )
        disposition = "attachment; filename=download.tar"
        resp['Content-Disposition'] = disposition

        return resp



from django.utils.module_loading import import_string

from django.core.exceptions import ImproperlyConfigured
from django.conf import settings


def get_search_backend(import_path=None):
    backend = import_path or settings.PAPERMERGE_SEARCH_BACKEND
    try:
        backend_cls = import_string(backend)
    except ImportError as e:
        raise ImproperlyConfigured("Could not find backend '%s': %s" % (
            backend, e))

    # Create backend
    return backend_cls()



from .registries import (
    navigation,
    user_menu_registry,
    sidebar
)

from .options import (
    SidebarPart,
    SidebarPartField
)

default_app_config = 'papermerge.contrib.admin.apps.AdminConfig'

__all__ = [
    'default_app_config',
    'navigation',
    'sidebar',
    'user_menu_registry',
    'SidebarPart',
    'SidebarPartField'
]



# Generated by Django 3.0.10 on 2020-10-02 18:14

from django.db import migrations
import papermerge.core.models.tags


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0024_auto_20200923_1412'),
    ]

    operations = [
        migrations.AlterField(
            model_name='automate',
            name='tags',
            field=papermerge.core.models.tags.UserTaggableManager(blank=True, help_text='A comma-separated list of tags.', through='core.ColoredTag', to='core.Tag', verbose_name='Tags'),
        ),
    ]



# Generated by Django 3.0.10 on 2020-11-18 23:32

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0027_auto_20201118_1538'),
    ]

    operations = [
        migrations.AlterField(
            model_name='automate',
            name='dst_folder',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='core.Folder', verbose_name='Destination Folder'),
        ),
    ]



# Generated by Django 3.0.2 on 2020-01-07 08:18

from django.conf import settings
import django.contrib.auth.models
import django.contrib.auth.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import papermerge.core.models
import polymorphic_tree.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0011_update_proxy_permissions'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('username', models.CharField(error_messages={'unique': 'A user with that username already exists.'}, help_text='Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.', max_length=150, unique=True, validators=[django.contrib.auth.validators.UnicodeUsernameValidator()], verbose_name='username')),
                ('first_name', models.CharField(blank=True, max_length=30, verbose_name='first name')),
                ('last_name', models.CharField(blank=True, max_length=150, verbose_name='last name')),
                ('email', models.EmailField(blank=True, max_length=254, verbose_name='email address')),
                ('is_staff', models.BooleanField(default=False, help_text='Designates whether the user can log into this admin site.', verbose_name='staff status')),
                ('is_active', models.BooleanField(default=True, help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.', verbose_name='active')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now, verbose_name='date joined')),
                ('current_storage_size', models.BigIntegerField(default=0)),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.Group', verbose_name='groups')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.Permission', verbose_name='user permissions')),
            ],
            options={
                'verbose_name': 'user',
                'verbose_name_plural': 'users',
                'abstract': False,
            },
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name='BaseTreeNode',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=200, verbose_name='Title')),
                ('lang', models.CharField(default='deu', max_length=8, verbose_name='Language')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('lft', models.PositiveIntegerField(editable=False)),
                ('rght', models.PositiveIntegerField(editable=False)),
                ('tree_id', models.PositiveIntegerField(db_index=True, editable=False)),
                ('level', models.PositiveIntegerField(editable=False)),
                ('parent', polymorphic_tree.models.PolymorphicTreeForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='core.BaseTreeNode', verbose_name='parent')),
                ('polymorphic_ctype', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_core.basetreenode_set+', to='contenttypes.ContentType')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Documents',
                'verbose_name_plural': 'Documents',
                'ordering': ('tree_id', 'lft'),
                'abstract': False,
                'base_manager_name': 'objects',
            },
        ),
        migrations.CreateModel(
            name='Document',
            fields=[
                ('basetreenode_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.BaseTreeNode')),
                ('file_orig', models.FileField(help_text='Reference to originaly imported file', max_length=512, upload_to='')),
                ('file_name', models.CharField(default='', max_length=1024)),
                ('notes', models.TextField(blank=True, null=True, verbose_name='Notes')),
                ('size', models.BigIntegerField(help_text='Size of file_orig attached. Size is in Bytes')),
                ('digest', models.CharField(blank=True, help_text='Digest of file_orig attached. Size is in Bytes.It is used to figure out if file was already processed.', max_length=512, null=True, unique=True)),
                ('page_count', models.IntegerField(default=1)),
                ('text', models.TextField()),
                ('celery_task_id', models.UUIDField(blank=True, default=None, null=True)),
            ],
            options={
                'verbose_name': 'Document',
                'verbose_name_plural': 'Documents',
            },
            bases=('core.basetreenode', ),
        ),
        migrations.CreateModel(
            name='Folder',
            fields=[
                ('basetreenode_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.BaseTreeNode')),
            ],
            options={
                'verbose_name': 'Folder',
                'verbose_name_plural': 'Folders',
            },
            bases=('core.basetreenode', ),
        ),
        migrations.CreateModel(
            name='Access',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('access_type', models.CharField(choices=[('allow', 'Allow'), ('deny', 'Deny')], default='allow', max_length=16)),
                ('access_inherited', models.BooleanField(default=False)),
                ('group', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='auth.Group', verbose_name='group')),
                ('node', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.BaseTreeNode')),
                ('permissions', models.ManyToManyField(to='auth.Permission')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'permissions': [('read', 'Read Access'), ('write', 'Write Access'), ('delete', 'Delete Access'), ('change_perm', 'Change Permissions Access'), ('take_ownership', 'Take Ownership Access')],
            },
        ),
        migrations.CreateModel(
            name='Page',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('number', models.IntegerField(default=1)),
                ('page_count', models.IntegerField(default=1)),
                ('text', models.TextField(default='')),
                ('lang', models.CharField(default='deu', max_length=8)),
                ('image', models.CharField(default='', max_length=1024)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                ('document', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Document')),
            ],
        ),
    ]



from .groups import (
    GroupsListView,
    GroupCreateView,
    GroupUpdateView
)
from .users import (
    UsersListView,
    UserCreateView,
    UserUpdateView,
    UserChangePasswordView
)
from .roles import (
    RolesListView,
    RoleCreateView,
    RoleUpdateView
)
from .tokens import (
    TokensListView,
    TokenCreateView,
    TokenUpdateView
)
from .tags import (
    TagsListView,
    TagCreateView,
    TagUpdateView
)
from .automates import (
    AutomatesListView,
    AutomateCreateView,
    AutomateUpdateView
)
from .logs import (
    LogsListView,
    LogUpdateView
)
from .index import (
    search,
    BrowseView,
    inbox_view,
    version_view
)
from .preferences import (
    preferences_view,
    preferences_section_view
)

__all__ = [
    'GroupsListView',
    'GroupCreateView',
    'GroupUpdateView',
    'UsersListView',
    'UserCreateView',
    'UserUpdateView',
    'UserChangePasswordView',
    'RolesListView',
    'RoleCreateView',
    'RoleUpdateView',
    'TokensListView',
    'TokenCreateView',
    'TokenUpdateView',
    'TagsListView',
    'TagCreateView',
    'TagUpdateView',
    'AutomatesListView',
    'AutomateCreateView',
    'AutomateUpdateView',
    'LogsListView',
    'LogUpdateView',
    'search',
    'BrowseView',
    'inbox_view',
    'version_view',
    'preferences_view',
    'preferences_section_view'
]



# Generated by Django 3.0.6 on 2020-06-05 10:42

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0013_auto_20200603_0724'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='kvstorepage',
            name='page_id',
        ),
        migrations.AddField(
            model_name='kvstorepage',
            name='page',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='kvstore', to='core.Page'),
        ),
    ]



# Generated by Django 3.0.8 on 2020-09-07 07:53

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import taggit.managers


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('core', '0019_auto_20200828_0847'),
    ]

    operations = [
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True, verbose_name='name')),
                ('slug', models.SlugField(max_length=100, unique=True, verbose_name='slug')),
                ('bg_color', models.CharField(default='00FF00', max_length=6)),
                ('fg_color', models.CharField(default='FFFFFF', max_length=6)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Tag',
                'verbose_name_plural': 'Tags',
            },
        ),
        migrations.CreateModel(
            name='ColoredTag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.IntegerField(db_index=True, verbose_name='object ID')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='core_coloredtag_tagged_items', to='contenttypes.ContentType', verbose_name='content type')),
                ('tag', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='core_coloredtag_items', to='core.Tag')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='basetreenode',
            name='tags',
            field=taggit.managers.TaggableManager(help_text='A comma-separated list of tags.', through='core.ColoredTag', to='core.Tag', verbose_name='Tags'),
        ),
    ]



import re

from django import template
from django.template import TemplateSyntaxError
from django.utils.safestring import mark_safe

register = template.Library()

# https://djangosnippets.org/snippets/661/


def _merge(lists):
    merged = []
    for words in lists:
        if merged:
            merged[-1] += words[0]
            del words[0]
        merged.extend(words)
    return merged


def highlight(
    text,
    phrases,
    class_name='success'
):
    if isinstance(phrases, str):
        phrases = [phrases]

    flags = re.I
    re_template = r"\b(%s)\b" or r"(%s)"
    expr = re.compile(re_template % "|".join(phrases), flags)
    template = '<span class="%s">%%s</span>' % class_name

    def replace(match):
        return template % match.group(0)

    highlighted = mark_safe(expr.sub(replace, text))

    return dict(
        original=text,
        highlighted=highlighted,
    )


def search_excerpt(
    text,
    phrases,
    context_words_count=5,
):
    if isinstance(phrases, str):
        phrases = [phrases]

    flags = re.I
    exprs = [
        re.compile("^%s$" % p, flags) for p in phrases
    ]
    whitespace = re.compile(r"\s+")
    re_template = r"\b(%s)\b" or r"(%s)"

    pattern = re_template % "|".join(phrases)
    pieces = re.compile(pattern, flags).split(text)
    matches = {}
    word_lists = []

    for i, piece in enumerate(pieces):
        word_lists.append(whitespace.split(piece))
        if i % 2:  # matched piece
            for expr in exprs:
                if expr.match(piece):
                    matches.setdefault(expr, []).append(i)

    i = 0
    merged = []
    for j in map(min, matches.values()):
        merged.append(
            _merge(word_lists[i:j])
        )
        merged.append(word_lists[j])
        i = j + 1

    merged.append(
        _merge(word_lists[i:])
    )

    output = []
    for i, words in enumerate(merged):
        omit = None
        if i == len(merged) - 1:
            omit = slice(max(1, 2 - i) * context_words_count + 1, None)
        elif i == 0:
            omit = slice(-context_words_count - 1)
        elif not i % 2:
            omit = slice(context_words_count + 1, -context_words_count - 1)
        if omit and words[omit]:
            words[omit] = ["..."]

        output.append(" ".join(words))

    return dict(
        original=text, excerpt="".join(output)
    )


class SearchExcerptNode(template.Node):

    def __init__(
        self,
        content,
        search_terms,
        context_words_count,
        var_name
    ):
        self._content = template.Variable(content)
        self._search_terms = template.Variable(search_terms)
        self._context_words_count = context_words_count
        self._var_name = var_name

    def render(self, context):
        try:
            content = self._content.resolve(context)
            search_terms = self._search_terms.resolve(context)
        except template.VariableDoesNotExist:
            return ''

        result = search_excerpt(
            text=content,
            phrases=search_terms,
            context_words_count=self._context_words_count
        )
        if self._var_name:
            context[self._var_name] = result
            return ""

        return result


class HighlightNode(template.Node):

    def __init__(
        self,
        content,
        search_terms,
        class_name,
        var_name
    ):
        self._content = template.Variable(content)
        self._search_terms = template.Variable(search_terms)
        self._class_name = class_name
        self._var_name = var_name

    def render(self, context):
        try:
            content = self._content.resolve(context)
            search_terms = self._search_terms.resolve(context)
        except template.VariableDoesNotExist:
            return ''

        result = highlight(
            text=content,
            phrases=search_terms,
            class_name=self._class_name
        )
        if self._var_name:
            context[self._var_name] = result
            return ""

        return result


@register.tag
def search_excerpt_tag(parser, token):
    """
        {% search_excerpt_tag content search_terms [word_count] [as name] %}
    """
    try:
        # Splitting by None == splitting by spaces.
        tag_name, arg = token.contents.split(None, 1)
    except ValueError:
        raise TemplateSyntaxError(
            "%r tag requires arguments" % token.contents.split()[0]
        )

    m = re.search(r'(.*?) as (\w+)', arg)
    if not m:
        raise TemplateSyntaxError(
            "%r tag had invalid arguments" % tag_name
        )

    before_var, var_name = m.groups()
    bits = before_var.split(None)

    if len(bits) > 2:
        context_words_count = bits[2]
    else:
        context_words_count = 5

    return SearchExcerptNode(
        content=bits[0],
        search_terms=bits[1],
        context_words_count=int(context_words_count),
        var_name=var_name
    )


@register.tag
def highlight_tag(parser, token):
    """
        {% highlight_tag text search_terms [class_name] [as name] %}
    """
    try:
        # Splitting by None == splitting by spaces.
        tag_name, arg = token.contents.split(None, 1)
    except ValueError:
        raise TemplateSyntaxError(
            "%r tag requires arguments" % token.contents.split()[0]
        )

    m = re.search(r'(.*?) as (\w+)', arg)
    if not m:
        raise TemplateSyntaxError(
            "%r tag had invalid arguments" % tag_name
        )

    before_var, var_name = m.groups()
    bits = before_var.split(None)

    if len(bits) > 2:
        class_name = bits[2]
    else:
        class_name = 'text-success'

    return HighlightNode(
        content=bits[0],
        search_terms=bits[1],
        class_name=class_name,
        var_name=var_name
    )



# Generated by Django 3.0.6 on 2020-06-03 06:07

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0011_auto_20200602_1532'),
    ]

    operations = [
        migrations.AddField(
            model_name='kvstore',
            name='kv_inherited',
            field=models.BooleanField(default=False),
        ),
        migrations.CreateModel(
            name='KVStoreCompNode',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('node', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='kvstorecomp', to='core.BaseTreeNode')),
            ],
        ),
        migrations.CreateModel(
            name='KVStoreCompItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('comp_item', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='kvstore', to='core.KVStore')),
            ],
        ),
    ]



# Generated by Django 3.0.2 on 2020-02-18 11:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0002_auto_20200116_1050'),
    ]

    operations = [
        migrations.AddField(
            model_name='document',
            name='version',
            field=models.IntegerField(blank=True, default=0, null=True),
        ),
    ]



from persisting_theory import Registry


class UserMenuRegistry(Registry):
    # the package where the registry will try to find callbacks in each app
    look_into = "user_menu"


class NavigationRegistry(Registry):
    """
    Leftside navigation menu registry.

    Leftside nagivation menu is where apps (parts of the system)
    appear.
    """

    # the package where the registry will look into
    look_into = "mg_admin"


class SidebarRegistry(Registry):
    look_into = "mg_admin"


user_menu_registry = UserMenuRegistry()  # noqa
navigation = NavigationRegistry()  # noqa
sidebar = SidebarRegistry()  # noqa



import logging

from django.conf import settings
from django.db import models
from django.utils.translation import gettext as _

from .diff import Diff
from papermerge.core.utils import (
    date_2int,
    money_2int,
    number_2int
)

"""
# KVStore / Key Value Store
####

 Implementation of metadata associated with Folders, Documents
 and Pages. Metadata can be defined at 3 levels:

    * Page
    * Document
    * Folder

If metadata is defined at Folder level - it will be inherited by all
descending nodes (subfolders and documents) and pages of respective
folders. Think of metadata definition as flowing down as a river.

Metadata is defined as a set of key names of specific type and format.
For example, metadata defined on folder Groceries could be:

    * key1 = date / type=date / format dd.mm.yy
    * key2 = shop / type=text / format=N/A
    * key3 = price / type=money / format dd,cc

This way, each and every document in Groceries folder (and respective pages)
will have associated those 3 keys. Other parts of application code will
populate the values for those keys.
As result, two page document (say D1) in folder Groceries will have:

 D1 content:
   page1 = key=date   value=03.06.2020
           key=shop   value=lidl
           key=price  value=34,02

   page2 = key=date   value=01.06.20
           key=shop   value=aldi
           key=price  value=19,00

 Which means D1 is one batch scan of two receipts. Other parts of application
 code might split page1 and page2 from D1 into two receipts documents with one
 page each.

 Groceries receitps are good example of simple KV instances.
 There are other type of KV instances - KVComp, or Key Value Composite.
 Consider a list of bank transactions. A content of document (DX)
 which contains bank transctions can be described as:

 DX content:

  page1 = key=(date, description, amount)  value=(01 June 2020, aldi, 19.00)
          key=(date, description, amount)  value=(03 June 2020, lidl, 34.02)

Basically DX document's metadata is one composite key. Thus, it would make
sense to create a folder titled say, Bank, create one composite key with 3
parts (keys):

  * compkey1 = (date, description, amount)

With compkey1 created for folder Bank, all folder's documents (and resp. pages)
will have associated multiple instances of compkey1.

There can be ONLY ONE composite key per Entity (E=Page, Node=Folder/Document)

KVComp describe tables. I repeat, as this is important -
only one table (compkey) per document is supported.
"""


logger = logging.getLogger(__name__)

# metadata types
TEXT = 'text'
MONEY = 'money'
NUMERIC = 'numeric'
DATE = 'date'

METADATA_TYPES = (
    (TEXT, _("Freeform Text")),
    (MONEY, _("Monetary")),
    (NUMERIC, _("Numeric")),
    (DATE, _("Date"))
)


def compute_virtual_value(kv_type, kv_format, value):
    """
    See comments for KVStore.virtual_value.
    In case of error, unknown kv_type, unkown kv_format,
    or error during conversion - log error and return 0.
    """
    result = 0
    if kv_type == DATE:
        result = date_2int(kv_format, value)
    elif kv_type == MONEY:
        result = money_2int(kv_format, value)
    elif kv_type == NUMERIC:
        result = number_2int(kv_format, value)
    elif kv_type == TEXT:
        # virtual_value is not used for kv_type=TEXT
        result = 0
    else:
        logger.error("Unknown kv_type value")
        return 0

    return result


def get_kv_types():
    return METADATA_TYPES


def get_currency_formats():
    return [
        (currency, _(currency))
        for currency in settings.PAPERMERGE_METADATA_CURRENCY_FORMATS
    ]


def get_numeric_formats():
    return [
        (numeric, _(numeric))
        for numeric in settings.PAPERMERGE_METADATA_NUMERIC_FORMATS
    ]


def get_date_formats():
    return [
        (_date, _(_date))
        for _date in settings.PAPERMERGE_METADATA_DATE_FORMATS
    ]


class KVCompKeyLengthMismatch(Exception):
    pass


class KVCompValidation(Exception):
    pass


class TypedKey:
    """ Used to compare inherited keys
    from parent to child (or from doc to page).
    Correcly inherited key is not just a key name - it is
    a combination of key, type and format.
    """

    def __init__(self, key, ktype, kformat):
        self.key = key
        self.ktype = ktype
        self.kformat = kformat

    def __eq__(self, other):

        _key = (self.key == other.key)
        _type = (self.ktype == other.ktype)
        _format = (self.kformat == other.kformat)

        return _key and _type and _format

    def __hash__(self):
        return hash((self.key, self.ktype, self.kformat))

    def __str__(self):
        return f"TypedKey({self.key}, {self.ktype}, {self.kformat})"

    def __repr__(self):
        return f"TypedKey({self.key}, {self.ktype}, {self.kformat})"


class KVComp:
    """
    Utility class that operates - adds, deletes, updates,
    values KVComp entries
    """

    def __init__(self, instance):
        self.instance = instance

    @property
    def namespace(self):
        pass

    def _validate(self, key, value):
        """
        Following validations are performed on the kvcomp:
            1. Both key and values must be instanced of either list of tuple
            2. len(key) > 0
            3. All rows (keys/values) must have same number of components.
            4. Column (key parts) names must match.
        """
        if not isinstance(key, (tuple, list)):
            raise KVCompValidation(
                "KVComp key must be a list or a tuple"
            )

        if not isinstance(value, (tuple, list)):
            raise KVCompValidation(
                "KVComp value must be a list or a tuple"
            )

        if len(key) == 0:
            raise KVCompValidation(
                "KVComp key must have more then 0 columns"
            )

        if self.instance.kvstorecomp.count() > 0:
            # there are other rows
            # compare agains first one
            row = self.instance.kvstorecomp.first()
            if row.kvstore.count() != len(key):
                raise KVCompKeyLengthMismatch(
                    "Existing column count does not match with new key length"
                )

            for kvstore in self.instance.kvstorecomp.all():
                if kvstore.key not in key:
                    k = kvstore.key
                    raise KVCompValidation(
                        f"Existing column name does not match for {k}"
                    )

    def all(self):
        result = []
        for row in self.instance.kvstorecomp.all():
            result.append(row.kvstore.all())

        return result

    def update(self, data):
        """
        data is one of:
            *
            *
        """
        pass

    def add(self, key, value=()):

        self._validate(key, value)

        # creates one row in the table
        new_row = self.instance.kvstorecomp.create()

        # add columns to the table
        for index, k in enumerate(key):
            if len(value) == len(key):
                new_row.kvstore.create(
                    namespace=self.namespace,
                    key=k,
                    value=value[index]
                )
            else:  # means value is empty, but key not
                new_row.kvstore.create(
                    namespace=self.namespace,
                    key=k,
                )

        new_row.save()


class KVCompNode(KVComp):
    pass


class KVCompPage(KVComp):
    pass


class KV:
    """
    Utility class that operates - adds, deletes, updates,
    values KV entries on KVStoreNode of Node (Folder or Document)
    and on KVStorePage of the Page
    """

    ADD = 'add'
    REMOVE = 'remove'
    UPDATE = 'update'

    class MetadataKeyDoesNotExist(Exception):
        """
        Raised when assigning (e.g. page.kv['shop'] = y)
        or retrieving a not existing metadata key - in example above - 'shop'
        """
        pass

    def __init__(self, instance):
        self.instance = instance

    @property
    def namespace(self):
        """
        returns namepace for added keys
        """
        pass

    def add(self, key):
        """
        adds a namespaced key,
        sends event signal for descendents to update themselves
        """
        instance = self.instance.kvstore.create(
            namespace=self.namespace,
            key=key
        )
        self.propagate(
            instances_set=[instance],
            operation=Diff.ADD
        )

    def __setitem__(self, key, value):

        try:
            instance = self.instance.kvstore.get(key=key)
        except Exception:
            raise KV.MetadataKeyDoesNotExist(
                f"Metadata key '{key}' does not exist."
            )

        if instance:
            instance.value = value
            instance.save()

    def __getitem__(self, key):
        try:
            instance = self.instance.kvstore.get(key=key)
        except Exception:
            raise KV.MetadataKeyDoesNotExist(
                f"Metadata key '{key}' does not exist."
            )

        if instance:
            return instance.value

        return None

    def keys(self):
        return [
            item.key for item in self.all()
        ]

    def typed_keys(self):
        return [
            item.to_typed_key() for item in self.all()
        ]

    def get_diff(self, data):
        result = {}
        result[KV.ADD] = []
        result[KV.REMOVE] = []
        result[KV.UPDATE] = []
        present_keys = self.keys()

        # discard empty string keys
        data = list(
            filter(lambda item: len(item['key']) > 0, data)
        )
        for item in data:
            if item['key'] not in present_keys:
                if item.get('id', False):
                    # key is not present, but it has an id => it is an update
                    result[KV.UPDATE].append(item)
                else:
                    # key is not present and it has no ID => it is new
                    result[KV.ADD].append(item)
            elif item.get('id', False):
                # presence of id attributes means an update.
                result[KV.UPDATE].append(item)

        # other way around (check for removes)
        data_keys = [item['key'] for item in data]
        for item in self.all():
            # if existing item's key is not found in data

            if item.key not in data_keys:
                # neither its id is found
                id_found = next(
                    filter(lambda x: x.get('id', False) == item.id, data),
                    False
                )
                # it means that user opted for key removal.
                if not id_found:
                    result[KV.REMOVE].append({
                        'key': item.key,
                        'id': item.id
                    })

        return result

    def apply_updates(self, updates):
        """
        updates is a list of dictionaries. Each dict is a will contain
        a key named "key" and one named "id" - the attributes of kvstore to be
        updated.
        """
        if len(updates) == 0:
            return

        attr_updates = []
        for item in updates:
            # update exiting
            if 'id' in item:
                kvstore_node = self.instance.kvstore.filter(
                    id=item['id']
                ).first()
                if kvstore_node:
                    # ok found it, just update the key
                    attr_updates.append({
                        # old key is needed in order to find the element
                        'old': kvstore_node.key,
                        'new': item['key'],
                        'kv_format': item['kv_format'],
                        'kv_type': item['kv_type'],
                        'value': item.get('value', None)
                    })
                    kvstore_node.key = item['key']
                    kvstore_node.kv_format = item['kv_format']
                    kvstore_node.kv_type = item['kv_type']
                    if item.get('value', False):
                        kvstore_node.value = item.get('value')
                    kvstore_node.save()
            elif 'old' in item and 'new' in item:
                kvstore_node = self.instance.kvstore.filter(
                    key=item['old']
                ).first()
                if kvstore_node:
                    # ok found it, just update the key
                    attr_updates.append({
                        'old': kvstore_node.key,
                        'new': item['new'],
                        'kv_format': item['kv_format'],
                        'kv_type': item['kv_type'],
                        'value': item.get('value', None)
                    })
                    kvstore_node.key = item['new']
                    kvstore_node.kv_format = item['kv_format']
                    kvstore_node.kv_type = item['kv_type']
                    if item.get('value', False):
                        kvstore_node.value = item.get('value')
                    kvstore_node.save()

        # It does not make sense to continue
        # propagation in case of pages.
        # Pages are leafs so to speak of hierarchy tree.
        # Thus propagation stops here.
        if isinstance(self, KVPage):
            return

        if updates:
            prop_updates = []
            for item in updates:
                if 'key' in item:
                    prop_updates.append(
                        KVStoreNode(
                            key=item['key'],
                            kv_type=item.get('kv_type', None),
                            kv_format=item.get('kv_format', None),
                        )
                    )
            self.propagate(
                instances_set=prop_updates,
                operation=Diff.UPDATE,
                attr_updates=attr_updates
            )

    def apply_additions(self, new_additions):
        """
        new_additions is a list of dictionaries. Each dict is a will contain
        a key named "key" - the key of kvstore to be added.
        """
        for item in new_additions:
            # look up by key
            kvstore_node = self.instance.kvstore.filter(
                key=item['key']
            ).first()
            # this key does not exist for this node
            if not kvstore_node:
                # delete a python id key regardless if it is present
                # or not.
                # The point here is that NEW element is created, so
                # presense of empty id key raises an error.
                item.pop('id', None)
                self.instance.kvstore.create(**item)
                self.instance.save()

        if new_additions:
            new_additions = [
                KVStoreNode(**item)
                for item in new_additions
            ]
            self.propagate(
                instances_set=new_additions,
                operation=Diff.ADD
            )

    def apply_deletions(self, deletions):
        """
        deletions is a list of dictionaries. Each dict is a will contain
        a key named "key" - the key of kvstore to be deleted.
        """
        for item in deletions:
            # look up by key
            kvstore_node = self.instance.kvstore.filter(
                key=item['key']
            ).first()
            self.instance.kvstore.remove(kvstore_node)

        if deletions:
            deletions = [
                KVStoreNode(key=item['key'])
                for item in deletions
            ]
            self.propagate(
                instances_set=deletions,
                operation=Diff.DELETE
            )

    def update(self, data):
        """
        data is one of:
            * list of dictionaries
        a dict can contain KVStore fields like:

            'key' = kvstore.key
            'id' = kvstore.id
        """

        if len(data) == 0:
            return

        kv_diff = self.get_diff(data)

        self.apply_updates(
            kv_diff[KV.UPDATE]
        )
        self.apply_additions(
            kv_diff[KV.ADD]
        )
        self.apply_deletions(
            kv_diff[KV.REMOVE]
        )

    def all(self):
        return self.instance.kvstore.all()

    def count(self):
        return self.instance.kvstore.count()

    def remove(self, key):
        """
        adds a namespaced key,
        sends event signal for descendents to update themselves
        """
        instances = self.instance.kvstore.filter(key=key)
        instances.delete()
        self.propagate(
            instances_set=[instances],
            operation=Diff.DELETE
        )

    def propagate(
        self,
        instances_set,
        operation,
        attr_updates=[]
    ):
        kv_diff = Diff(
            operation=operation,
            instances_set=instances_set
        )
        self.instance.propagate_changes(
            diffs_set=[kv_diff],
            apply_to_self=False,
            attr_updates=attr_updates
        )


class KVNode(KV):
    """
    KV specific to Node (Folder or Document)
    """

    def __eq__(self, kvnode):
        """
        Two KVNodes are equal if all conditions
        are true:
        * kvnode is attached to the same node
        * have name number of keys
        * keys names are same
        """
        if self.kv.instance.id != kvnode.instance.id:
            return False

        if len(self.kv.all()) != len(kvnode.instance.kv.all()):
            return False

        keys_set1 = set(self.kv.keys())
        keys_set2 = set(kvnode.kv.keys())
        if keys_set1 != keys_set2:
            return False

        return True


class KVPage(KV):
    """
    KV specific to Page
    """

    def apply_additions(self, new_additions):
        """
        new_additions is a list of dictionaries. Each dict is a will contain
        a key named "key" - the key of kvstore to be added.
        """
        for item in new_additions:
            # look up by key
            kvstore_node = self.instance.kvstore.filter(
                key=item['key']
            ).first()
            # this key does not exist for this node
            if not kvstore_node:
                self.instance.kvstore.create(**item)


class KVStore(models.Model):
    # e.g. shop, price, date
    key = models.CharField(
        max_length=200,
        null=True,
        blank=True
    )
    # e.g. groceries - full breadcrumb to the current node
    namespace = models.CharField(
        max_length=1024,
        null=True,
        blank=True
    )
    # (namespace, key) = primary key
    # e.g.
    # 1.   groceries_shop
    #      groceries_price
    # 2. key = shop
    #    namespace = clients_retail
    #    namespaced_key = clients_retail_shop
    namespaced_key = models.CharField(
        max_length=200,
        null=True,
        blank=True
    )

    value = models.TextField(
        null=True,
        blank=True
    )

    kv_type = models.CharField(
        max_length=16,
        choices=[
            (TEXT, _('Text')),
            (MONEY, _('Currency')),
            (NUMERIC, _('Numeric')),
            (DATE, _('Date'))
        ],
        default='text'
    )

    # Format for kv_type
    #
    # e.g. for currency => dd.cc or may be dd,cc
    #      for date => dd.mm.yy or dd Month YYYY
    #      for numeric => 1,2000 or 1200 or 1.200
    #
    # available formats are loaded from configuration file
    # and will be displayed for user to choose from when
    # creating/adding a metadata.
    #
    # Without specifing format - metadata is basically useless.
    # what does value=05.06.20 key=date mean ?
    # is it 5th of June 2020 or it is 6th of May 2020 ?
    kv_format = models.CharField(
        max_length=64,
        null=True,
        blank=True
    )

    # inherited kv is read only
    kv_inherited = models.BooleanField(
        default=False
    )

    def to_dict(self):
        item = {}
        item['id'] = self.id
        item['key'] = self.key
        item['kv_type'] = self.kv_type
        item['kv_format'] = self.kv_format
        item['kv_inherited'] = self.kv_inherited
        item['value'] = self.value
        item['virtual_value'] = self.virtual_value
        item['kv_types'] = get_kv_types()
        item['currency_formats'] = get_currency_formats()
        item['numeric_formats'] = get_numeric_formats()
        item['date_formats'] = get_date_formats()

        return item

    def to_typed_key(self):

        return TypedKey(
            key=self.key,
            ktype=self.kv_type,
            kformat=self.kv_format
        )

    @property
    def virtual_value(self):
        """
        An integer number e.g. 1, 100, 345 which is used for sorting.
        It is deduced from 3 components: (value, kv_type, kv_format).

        Example:
        given two metadata items:

            m1 = (date, 'dd.mm.yy', '04.05.20')
            m2 = (date, 'dd.mm.yy', '05.06.20')

        How to order them? Well, here is where virtual_value comes into
        the picture. Virtual value is computed, as integer number,
        and based on it, m1 and m2 are ordered.
        """
        result = compute_virtual_value(
            self.kv_type,
            self.kv_format,
            self.value
        )
        return result


class KVStoreNode(KVStore):
    node = models.ForeignKey(
        'BaseTreeNode',
        on_delete=models.CASCADE,
        related_name='kvstore',
        null=True
    )

    def __str__(self):
        k = self.key
        v = self.value
        if self.node:
            n = self.node.id
        else:
            n = None
        s = self.namespace
        f = self.kv_format
        t = self.kv_type

        return f"KVStoreNode(namespace={s}," + \
            f" key={k}, kv_type={t}, kv_format={f}, value={v}, node={n})"

    def __repr__(self):
        return str(self)


class KVStorePage(KVStore):
    """
    Normalized KVStore per Page
    """
    page = models.ForeignKey(
        'Page',
        on_delete=models.CASCADE,
        related_name='kvstore',
        null=True
    )

    def __str__(self):
        k = self.key
        v = self.value
        p = self.page.id
        s = self.namespace
        f = self.kv_format
        t = self.kv_type

        return f"KVStorePage(namespace={s}," + \
            f" key={k}, kv_type={t}, kv_format={f} , value={v}, page={p})"

    def __repr__(self):
        return str(self)


class KVStoreCompItem(KVStore):
    comp_item = models.ForeignKey(
        'KVStoreCompNode',
        on_delete=models.CASCADE,
        related_name='kvstore',
        null=True,
        blank=True
    )

    def __str__(self):
        k = self.key
        v = self.value
        s = self.namespace
        return f"KVStoreCompItem(namespace={s}, key={k}, value={v})"

    def __repre__(self):
        return str(self)


class KVStoreCompNode(models.Model):
    node = models.ForeignKey(
        'BaseTreeNode',
        on_delete=models.CASCADE,
        related_name='kvstorecomp',
        null=True,
        blank=True
    )



import logging
import os

from django.db import models

from mglib.path import PagePath
from papermerge.core.storage import default_storage
from papermerge.search import index
from papermerge.search.queryset import SearchableQuerySetMixin

from .diff import Diff
from .document import Document
from .kvstore import KVCompPage, KVPage, KVStorePage

logger = logging.getLogger(__name__)


class PageQuerySet(SearchableQuerySetMixin, models.QuerySet):
    pass


class Page(models.Model, index.Indexed):
    document = models.ForeignKey(
        Document,
        on_delete=models.CASCADE,
        related_name='pages'
    )
    user = models.ForeignKey('User', models.CASCADE)

    number = models.IntegerField(default=1)
    page_count = models.IntegerField(default=1)

    text = models.TextField(default='')

    # inherited/normalized title from parent document
    norm_doc_title = models.CharField(
        max_length=200,
        default=''
    )

    # inherited/normalized title of immediate parent folder
    norm_folder_title = models.CharField(
        max_length=200,
        default=''
    )

    # normalized space delimited path (by folder title) of parent folder
    norm_breadcrump = models.CharField(
        max_length=1024,
        default=''
    )

    # text from all pages of the document
    norm_text = models.TextField(default='')

    # hm, this one should be norm_lang as well
    lang = models.CharField(
        max_length=8,
        blank=False,
        null=False,
        default='deu'
    )

    search_fields = [
        index.SearchField('norm_doc_title', partial_match=True, boost=3),
        index.SearchField('norm_folder_title', partial_match=True),
        index.SearchField('norm_breadcrump', partial_match=True),
        index.SearchField('norm_text', partial_match=True, boost=1),
        index.SearchField('text', partial_match=True, boost=2),
        index.FilterField('lang')
    ]

    image = models.CharField(
        max_length=1024,
        default=''
    )
    # The hocr (text) fields corresponding
    # to step=0, step=1, ..., step=3 will be saved in hocr_step_0, ...,
    # hocr_step_3 fields.
    #
    # Read in header comment of python module provided by URL below about
    # the concept of steps and why it was bad design
    # decision to introduce them:
    #
    #   https://github.com/papermerge/mglib/blob/master/mglib/step.py
    #
    # In future releases 4 fields: hocr_step_0,..., hocr_step_3 will be
    # replaced with just one: hocr = models.TextField(default='')
    hocr_step_0 = models.TextField(
        default=''
    )
    hocr_step_1 = models.TextField(
        default=''
    )
    hocr_step_2 = models.TextField(
        default=''
    )
    hocr_step_3 = models.TextField(
        default=''
    )

    objects = PageQuerySet.as_manager()

    def to_dict(self):

        item = {}
        item['id'] = self.id
        item['number'] = self.number
        item['kvstore'] = [item.to_dict() for item in self.kv.all()]

        return item

    @property
    def kv(self):
        return KVPage(instance=self)

    @property
    def kvcomp(self):
        return KVCompPage(instance=self)

    def _apply_diff_add(self, diff):

        self.kv.apply_additions(
            [
                {
                    'kv_inherited': True,
                    'key': _model.key,
                    'kv_format': _model.kv_format,
                    'kv_type': _model.kv_type
                }
                for _model in diff
            ]
        )

    def _apply_diff_update(self, diff, attr_updates):
        updates = [{
            'kv_inherited': True,
            'key': _model.key,
            'kv_format': _model.kv_format,
            'kv_type': _model.kv_type,
            'id': _model.id
        } for _model in diff]

        updates.extend(attr_updates)

        self.kv.apply_updates(updates)

    def _apply_diff_delete(self, diff):
        pass

    def apply_diff(self, diffs_list, attr_updates):

        for diff in diffs_list:
            if diff.is_add():
                self._apply_diff_add(diff)
            elif diff.is_update():
                self._apply_diff_update(diff, attr_updates)
            elif diff.is_delete():
                self._apply_diff_delete(diff)
            elif diff.is_replace():
                # not applicable to page model
                # replace is used in access permissions
                # propagation
                pass
            else:
                raise ValueError(
                    f"Unexpected diff {diff} type"
                )

    def inherit_kv_from(self, document):
        instances_set = []

        for kvstore in document.kv.all():
            instances_set.append(
                KVStorePage(
                    key=kvstore.key,
                    kv_format=kvstore.kv_format,
                    kv_type=kvstore.kv_type,
                    value=kvstore.value,
                    kv_inherited=True,
                    page=self
                )
            )

        diff = Diff(
            operation=Diff.ADD,
            instances_set=instances_set
        )

        self.propagate_changes(
            diffs_set=[diff],
        )

    def propagate_changes(
        self,
        diffs_set,
        apply_to_self=None,
        attr_updates=[]
    ):
        """
        apply_to_self argument does not make sense here.
        apply_to_self argument is here to make this function
        similar to node.propagate_changes.
        """
        self.apply_diff(
            diffs_list=diffs_set,
            attr_updates=attr_updates
        )

    @property
    def is_last(self):
        return self.number == self.page_count

    @property
    def is_first(self):
        return self.number == 1

    def path(self, version=None):

        return PagePath(
            document_path=self.document.path(version=version),
            page_num=self.number,
            page_count=self.page_count
        )

    def update_text_field(self):
        """Update text field from associated .txt file.

        Returns non-empty text string value if .txt file was found.
        If file was not found - will return an empty string.
        """
        text = ''
        url = default_storage.abspath(self.txt_url)

        if not os.path.exists(url):
            logger.debug(
                f"Missing page txt {url}."
            )
            return

        with open(url) as file_handle:
            self.text = file_handle.read()
            self.save()
            logger.debug(
                f"text saved. len(page.text)=={len(self.text)}"
            )
            text = self.text

        return text

    @property
    def txt_url(self):

        result = PagePath(
            document_path=self.document.path(),
            page_num=self.number,
            page_count=self.page_count
        )

        return result.txt_url()

    @property
    def txt_exists(self):

        result = PagePath(
            document_path=self.document.path(),
            page_num=self.number,
            page_count=self.page_count
        )

        return result.txt_exists()

    def norm(self):
        """shortcut normalization method"""
        self.normalize_doc_title()
        self.normalize_folder_title()
        self.normalize_breadcrump()
        self.normalize_text()
        self.normalize_lang()

    def normalize_doc_title(self):
        """
        Save containing document's title
        """
        self.norm_doc_title = self.document.title
        self.save()

    def normalize_folder_title(self):
        """
        Save direct parent folder (containing folder) title
        """
        if self.document.parent:
            self.norm_folder_title = self.document.parent.title
            self.save()

    def normalize_breadcrump(self):
        pass

    def normalize_text(self):
        pass

    def normalize_lang(self):
        pass

    class Meta:
        # Guarantees that
        # doc.pages.all() will return pages ordered by number.
        # test by
        # test_page.TestPage.test_pages_all_returns_pages_ordered
        ordering = ['number']


def get_pages(
    nodes,
    include_pages_with_empty_text=True
):
    """
    :nodes: is a query set of BaseTreeNodes

    Returns a QuerySet of Page(s)
    """
    if nodes.count() == 0:
        return Page.objects.none()

    pages = Page.objects.filter(
        document__in=nodes
    )

    if not include_pages_with_empty_text:
        ret_pages = pages.exclude(
            text__isnull=True
        ).exclude(
            text__exact=''
        )
    else:
        ret_pages = pages

    return ret_pages



from datetime import timedelta

from django.contrib import messages
from django.http import HttpResponseRedirect
from django.utils.translation import ugettext_lazy as _
from django.views import generic
from django.urls import reverse_lazy
from django.contrib.auth.mixins import LoginRequiredMixin

from knox.models import AuthToken

from papermerge.contrib.admin.forms import AuthTokenForm
from papermerge.contrib.admin.views import mixins as mix


class TokensView(
    LoginRequiredMixin,
    mix.CommonListMixin
):
    model = AuthToken
    form_class = AuthTokenForm
    success_url = action_url = reverse_lazy('admin:tokens')
    new_object_url = reverse_lazy('admin:token-add')


class TokensListView(
    TokensView,
    mix.PaginationMixin,
    mix.DeleteEntriesMixin,
    generic.ListView
):
    title = _("Tokens")
    action_url = reverse_lazy('admin:tokens')

    table_header_row = [
        _('Digest'),
        _('Created At'),
        _('Expires At'),
    ]

    def get_queryset(self):
        qs = super().get_queryset()

        return qs.filter(user=self.request.user)

    def get_delete_entries(self, selection):
        return self.get_queryset().filter(
            digest__in=selection
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['table_header_row'] = self.table_header_row
        return context


class TokenCreateView(TokensView, generic.CreateView):

    title = _("New")
    action_url = reverse_lazy('admin:token-add')
    template_name = "admin/object_form.html"

    def form_valid(self, form):

        instance, token = AuthToken.objects.create(
            user=self.request.user,
            expiry=timedelta(hours=form.cleaned_data['hours'])
        )
        html_message = f"Please remember the token: {token}"
        html_message += " It won't be displayed again."

        messages.success(
            self.request, html_message
        )
        return HttpResponseRedirect(self.success_url)


class TokenUpdateView(TokensView, generic.UpdateView):

    template_name = "admin/object_form.html"

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)
        context['title'] = _('Edit')
        context['action_url'] = reverse_lazy(
            'admin:token-update',
            args=(self.object.pk,)
        )

        return context




import re
import os

SUPPORTED_EXTENTIONS = re.compile(".*(jpeg|jpg|png|tiff|pdf)$", re.IGNORECASE)


def filter_by_extention(
    names_list,
    compiled_reg_expr=SUPPORTED_EXTENTIONS
):
    """
    input:
        names_list
            is a list/tubple with string entries.
        compiled_reg_expr
            compiled regular expression pattern for file extentions to filter.
            Only file names of whom extentions match compiled_reg_expr
            will stay in returned list

    returns another list containing only entries with extention
    matching given pattern.
    """
    result = []

    for name in names_list:
        root, ext = os.path.splitext(name)
        if compiled_reg_expr.match(ext):
            result.append(name)

    return result



from django.urls import path
from papermerge.contrib.admin import views

app_name = 'admin'

urlpatterns = [
    path(
        '', views.BrowseView.as_view(), name="index"
    ),
    path(
        'inbox/', views.inbox_view, name="inbox"
    ),
    path('version/', views.version_view, name="version"),
    path(
        'browse', views.BrowseView.as_view(), name="browse"
    ),
    path(
        'search', views.search, name="search"
    ),
    path(
        'logs', views.LogsListView.as_view(), name="logs"
    ),

    path(
        'log/<int:pk>/',
        views.LogUpdateView.as_view(),
        name="log-update"
    ),
    path(
        'tags/', views.TagsListView.as_view(), name="tags"
    ),
    path(
        'tag/add/',
        views.TagCreateView.as_view(),
        name="tag-add"
    ),
    path(
        'tag/<int:pk>/',
        views.TagUpdateView.as_view(),
        name='tag-update'
    ),
    path(
        'groups/',
        views.GroupsListView.as_view(),
        name='groups'
    ),
    path(
        'group/add/',
        views.GroupCreateView.as_view(),
        name='group-add'
    ),
    path(
        'group/<int:pk>/',
        views.GroupUpdateView.as_view(),
        name='group-update'
    ),
    path(
        'users/',
        views.UsersListView.as_view(),
        name='users'
    ),
    path(
        'user/add/',
        views.UserCreateView.as_view(),
        name='user-add'
    ),
    path(
        'user/<int:pk>/',
        views.UserUpdateView.as_view(),
        name='user-update'
    ),
    path(
        'user/<int:pk>/change-password',
        views.UserChangePasswordView.as_view(),
        name='user-change-password'
    ),
    path(
        'roles/',
        views.RolesListView.as_view(),
        name='roles'
    ),
    path(
        'role/add/',
        views.RoleCreateView.as_view(),
        name='role-add'
    ),
    path(
        'role/<int:pk>/',
        views.RoleUpdateView.as_view(),
        name='role-update'
    ),
    path(
        'tokens/',
        views.TokensListView.as_view(),
        name='tokens'
    ),
    path(
        'token/add/',
        views.TokenCreateView.as_view(),
        name='token-add'
    ),
    path(
        'token/<int:pk>/',
        views.TokenUpdateView.as_view(),
        name='token-update'
    ),
    path(
        'automates/',
        views.AutomatesListView.as_view(),
        name='automates'
    ),
    path(
        'automate/add/',
        views.AutomateCreateView.as_view(),
        name='automate-add'
    ),
    path(
        'automate/<int:pk>',
        views.AutomateUpdateView.as_view(),
        name='automate-update'
    ),
    path(
        'preferences/',
        views.preferences_view,
        name='preferences'
    ),
    path(
        'preferences/<str:section>/',
        views.preferences_section_view,
        name='preferences_section'
    ),
]



# Generated by Django 3.1.5 on 2021-01-17 13:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0035_auto_20201207_1137'),
    ]

    operations = [
        migrations.AlterField(
            model_name='document',
            name='file_name',
            field=models.CharField(default='', max_length=1024),
        ),
        migrations.AlterField(
            model_name='document',
            name='notes',
            field=models.TextField(blank=True, null=True, verbose_name='Notes'),
        ),
    ]



from django.contrib.auth.models import Group
from django.utils.translation import ugettext_lazy as _
from django.views import generic
from django.urls import reverse_lazy
from django.contrib.auth.mixins import LoginRequiredMixin

from papermerge.contrib.admin.forms import GroupForm
from papermerge.contrib.admin.views import mixins as mix


class GroupsView(
    LoginRequiredMixin,
    mix.CommonListMixin,
    mix.RequiredPermissionMixin
):
    # only superuser can access this view
    only_superuser = True
    model = Group
    form_class = GroupForm
    success_url = action_url = reverse_lazy('admin:groups')
    new_object_url = reverse_lazy('admin:group-add')


class GroupsListView(
    GroupsView,
    mix.PaginationMixin,
    mix.DeleteEntriesMixin,
    generic.ListView
):
    required_permission = 'auth.view_group'
    title = _("Groups")
    table_header_row = [
        _('Name'),
    ]

    def get_queryset(self):
        qs = super().get_queryset()
        return qs.order_by('name')

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['table_header_row'] = self.table_header_row
        return context


class GroupCreateView(GroupsView, generic.CreateView):

    required_permission = 'auth.add_group'
    title = _("New")
    action_url = reverse_lazy('admin:group-add')
    template_name = "admin/object_form.html"


class GroupUpdateView(GroupsView, generic.UpdateView):

    required_permission = 'auth.change_group'
    title = _("Edit")
    template_name = "admin/object_form.html"

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)

        context['action_url'] = reverse_lazy(
            'admin:group-update',
            args=(self.object.pk,)
        )

        return context



import json
import logging

from django.contrib.auth.decorators import login_required
from django.http import (
    HttpResponseBadRequest,
    HttpResponseForbidden
)
from django.utils.translation import gettext as _
from django.core.exceptions import ValidationError

from papermerge.core import validators
from papermerge.core.models import Access, BaseTreeNode, Tag
from .decorators import json_response

logger = logging.getLogger(__name__)


@json_response
@login_required
def tags_view(request, node_id):
    """
    Add/remove tags per single node
    """
    try:
        node = BaseTreeNode.objects.get(id=node_id)
    except BaseTreeNode.DoesNotExist:
        msg = _("Node does not exist")
        return msg, HttpResponseBadRequest.status_code

    if request.user.has_perm(Access.PERM_WRITE, node):
        data = json.loads(request.body)
        tags = [item['name'] for item in data['tags']]
        try:
            # validate user's input
            for tag in tags:
                validators.safe_character_validator(tag)
        except ValidationError as e:
            return e.message, HttpResponseBadRequest.status_code

        node.tags.set(
            *tags,
            tag_kwargs={"user": request.user}
        )
    else:
        msg = _(
            "%(username)s does not have "
            "write permission on node %(title)s"
        ) % {
            'username': request.user.username,
            'title': node.title
        }

        return msg, HttpResponseForbidden.status_code

    return 'OK'


@json_response
@login_required
def nodes_tags_view(request):
    """
    Add/remove tags per bulk of nodes.

    Example of valid input:
        {"tags":[{"name":"test"}],"nodes":[92,52]}

    There must be at least one node, otherwise
    HttpReponseBadRequest will be returned.
    Tags however, can be empty. In case of empty tag list -
    all given nodes will be "stripped" from all tags - sort
    of bulk remove.
    """
    data = json.loads(request.body)
    node_ids = data.get('nodes', [])
    tags = data.get('tags', [])

    if not node_ids:
        msg = _("No nodes provided")
        return msg, HttpResponseBadRequest.status_code

    tags = [item['name'] for item in tags]

    nodes = BaseTreeNode.objects.filter(id__in=node_ids)

    nodes_perms = request.user.get_perms_dict(
        nodes, Access.ALL_PERMS
    )

    for node in nodes:
        if nodes_perms[node.id].get(Access.PERM_WRITE, False):
            try:
                # validate user's input
                for tag in tags:
                    validators.safe_character_validator(tag)
            except ValidationError as e:
                return e.message, HttpResponseBadRequest.status_code

            node.tags.set(
                *tags,
                tag_kwargs={"user": request.user}
            )
        else:
            msg = _(
                "%(username)s does not have "
                "write permission on node %(title)s"
            ) % {
                'username': request.user.username,
                'title': node.title
            }

            return msg, HttpResponseForbidden.status_code

    return 'OK'


@json_response
@login_required
def alltags_view(request):

    all_tags = [
        {'name': tag.name}
        for tag in Tag.objects.filter(user=request.user)
    ]

    return {'tags': all_tags}



# Generated by Django 3.0.10 on 2020-10-18 08:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0026_auto_20201113_1427'),
    ]

    operations = [
        migrations.AlterField(
            model_name='tag',
            name='name',
            field=models.CharField(max_length=100, verbose_name='name'),
        ),
        migrations.AlterUniqueTogether(
            name='tag',
            unique_together={('name', 'user')},
        ),
    ]



# Generated by Django 3.0.6 on 2020-06-03 07:24

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0012_auto_20200603_0607'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='kvstorecompitem',
            name='id',
        ),
        migrations.AddField(
            model_name='kvstorecompitem',
            name='kvstore_ptr',
            field=models.OneToOneField(auto_created=True, default='', on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.KVStore'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='kvstorecompitem',
            name='comp_item',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='kvstore', to='core.KVStoreCompNode'),
        ),
        migrations.AlterField(
            model_name='kvstorecompnode',
            name='node',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='kvstorecomp', to='core.BaseTreeNode'),
        ),
    ]



import os
import json
import logging

from django.utils.translation import gettext as _
from django.utils.html import escape
from django.shortcuts import redirect, render, get_object_or_404
from django.urls import reverse
from django.views.decorators.http import require_POST
from django.conf import settings
from django.contrib import messages
from django.http import (
    HttpResponse,
    HttpResponseRedirect,
    HttpResponseBadRequest,
    HttpResponseForbidden,
    Http404
)
from django.core.exceptions import ValidationError
from django.contrib.staticfiles import finders
from django.contrib.auth.decorators import login_required

from mglib.pdfinfo import get_pagecount
from mglib.step import Step
from mglib.shortcuts import extract_img

from papermerge.core.storage import default_storage
from papermerge.core.lib.hocr import Hocr
from .decorators import json_response, require_PERM

from papermerge.core.models import (
    Folder,
    Document,
    Page,
    BaseTreeNode,
    Access
)
from papermerge.core.utils import filter_node_id
from papermerge.core import signal_definitions as signals
from papermerge.core.import_pipeline import WEB, go_through_pipelines
from papermerge.core.tasks import ocr_page

logger = logging.getLogger(__name__)


@login_required
def document(request, doc_id):
    try:
        doc = Document.objects.get(id=doc_id)
    except Document.DoesNotExist:
        return render(request, "admin/document_404.html")

    nodes_perms = request.user.get_perms_dict(
        [doc], Access.ALL_PERMS
    )

    version = request.GET.get('version', None)
    if version is not None:
        version = int(version)

    if doc.is_latest_version(version):
        pagelist = doc.pages.all()
    else:
        pagelist = [{
            'id': page_num,
            'number': page_num
        } for page_num in range(1, doc.get_pagecount(version=version) + 1)]

    # request.is_ajax is repricated since Django 3.1
    if not (request.headers.get('x-requested-with') == 'XMLHttpRequest'):
        if request.user.has_perm(Access.PERM_READ, doc):
            if not doc.is_latest_version(version):
                messages.info(
                    request, _(
                        "This is past version of the document."
                        " Content of this version is read only."
                    )
                )
            return render(
                request,
                'admin/document.html',
                {
                    'pages': pagelist,
                    'tags': doc.tags.all(),
                    'document': doc,
                    'versions': doc.get_versions(),
                    'is_latest_version': doc.is_latest_version(version),
                    'version': version,
                    'has_perm_write': nodes_perms[doc.id].get(
                        Access.PERM_WRITE, False
                    ),
                }
            )
        else:
            return HttpResponseForbidden()

    # ajax + PATCH
    if request.method == 'PATCH':
        # test_document_view
        # TestDocumentAjaxOperationsView.test_update_notes
        # test_update_notes
        if request.user.has_perm(Access.PERM_WRITE, doc):
            data = json.loads(request.body)
            if 'notes' in data:
                # dangerous user input. Escape it.
                doc.notes = escape(data['notes'])
                doc.save()
                return HttpResponse(
                    json.dumps(
                        {
                            'msg': _("Notes saved!")
                        }
                    ),
                    content_type="application/json",
                )
        else:
            return HttpResponseForbidden(
                json.dumps({'msg': _("Access denied")}),
                content_type="application/json",
            )

    if request.method == 'DELETE':
        # test_document_view
        # TestDocumentAjaxOperationsView.test_delete_document
        if request.user.has_perm(Access.PERM_DELETE, doc):
            doc.delete()
            return HttpResponse(
                json.dumps(
                    {
                        'msg': "OK",
                        'url': reverse('admin:browse')
                    }
                ),
                content_type="application/json",
            )
        else:
            return HttpResponseForbidden(
                json.dumps(_("Access denied")),
                content_type="application/json",
            )

    # so, ajax only here
    if request.method == 'POST':
        # ajax + post
        pass

    # ajax + GET here
    result_dict = doc.to_dict()
    result_dict['user_perms'] = nodes_perms[doc.id]

    response = HttpResponse(
        json.dumps({'document': result_dict}),
        content_type="application/json",
    )

    response['Vary'] = 'Accept'

    return response


@json_response
@login_required
@require_POST
def cut_node(request):
    data = json.loads(request.body)
    node_ids = [item['id'] for item in data]
    nodes = BaseTreeNode.objects.filter(
        id__in=node_ids
    )
    nodes_perms = request.user.get_perms_dict(
        nodes, Access.ALL_PERMS
    )
    for node in nodes:
        if not nodes_perms[node.id].get(
            Access.PERM_DELETE, False
        ):
            msg = _(
                "%s does not have permission to cut %s"
            ) % (request.user.username, node.title)

            return msg, HttpResponseForbidden.status_code

    # request.clipboard.nodes = request.nodes
    request.nodes.add(node_ids)

    return 'OK'


@login_required
def clipboard(request):
    if request.method == 'GET':

        return HttpResponse(
            # request.nodes = request.clipboard.nodes
            json.dumps({'clipboard': request.nodes.all()}),
            content_type="application/json",
        )

    return HttpResponse(
        json.dumps({'clipboard': []}),
        content_type="application/json",
    )


@login_required
@require_POST
def paste_pages(request):
    """
    Paste pages in a changelist view.
    This means a new document instance
    is created.
    """
    data = json.loads(request.body)
    parent_id = data.get('parent_id', None)

    if parent_id:
        parent_id = int(parent_id)

    Document.paste_pages(
        user=request.user,
        parent_id=parent_id,
        doc_pages=request.pages.all()
    )

    request.pages.clear()

    return HttpResponse(
        json.dumps({'msg': 'OK'}),
        content_type="application/json",
    )


@login_required
@require_POST
def paste_node(request):
    data = json.loads(request.body)

    if not data:
        return HttpResponseBadRequest(
            json.dumps({
                'msg': 'Payload empty'
            }),
            content_type="application/json"
        )

    parent_id = data.get('parent_id', False)

    if parent_id:
        parent = BaseTreeNode.objects.filter(id=parent_id).first()
    else:
        parent = None

    # request.clipboard.nodes = request.nodes
    node_ids = request.nodes.all()

    # iterate through all node ids and change their
    # parent to new one (parent_id)
    for node in BaseTreeNode.objects.filter(id__in=node_ids):
        node.refresh_from_db()
        if parent:
            parent.refresh_from_db()
        Document.objects.move_node(node, parent)

    # request.clipboard.nodes = request.nodes
    request.nodes.clear()

    return HttpResponse(
        json.dumps({
            'msg': 'OK'
        }),
        content_type="application/json"
    )


@json_response
@login_required
def rename_node(request, id):
    """
    Renames a node (changes its title field).
    """

    data = json.loads(request.body)
    title = data.get('title', None)
    node = get_object_or_404(BaseTreeNode, id=id)

    if not request.user.has_perm(Access.PERM_WRITE, node):
        msg = _(
            "You don't have permissions to rename this document"
        )
        return msg, HttpResponseForbidden.status_code

    if not title:
        return _('Missing title')

    node.title = title
    # never trust data coming from user
    try:
        node.full_clean()
    except ValidationError as e:
        return e.message_dict, HttpResponseBadRequest.status_code

    node.save()

    return 'OK'


@login_required
@require_POST
@require_PERM('core.add_folder')
def create_folder(request):
    """
    Creates a new folder.

    Mandatory parameters parent_id and title:
    * If either parent_id or title are missing - does nothing.
    * If parent_id < 0 => creates a folder with parent root.
    * If parent_id >= 0 => creates a folder with given parent id.
    """
    data = json.loads(request.body)
    parent_id = data.get('parent_id', -1)
    title = data.get('title', False)

    if title == Folder.INBOX_NAME:
        return HttpResponseBadRequest(
            json.dumps({
                'msg': 'This title is not allowed'
            }),
            content_type="application/json"
        )

    if not (parent_id or title):
        logger.info(
            "Invalid params for create_folder: parent=%s title=%s",
            parent_id,
            title
        )
        return HttpResponseBadRequest(
            json.dumps({
                'msg': 'Both parent_id and title empty'
            }),
            content_type="application/json"
        )
    try:
        parent_id = int(parent_id or -1)
    except ValueError:
        parent_id = -1

    if int(parent_id) < 0:
        parent_folder = None
    else:
        parent_folder = Folder.objects.filter(id=parent_id).first()
        # if not existing parent_id was given, redirect to root
        if not parent_folder:
            return HttpResponseBadRequest(
                json.dumps({
                    'msg': f"Parent with id={parent_id} does not exist"
                }),
                content_type="application/json"
            )
    folder = Folder(
        title=title,
        parent=parent_folder,
        user=request.user
    )
    try:
        folder.full_clean()
    except ValidationError as e:
        # create human friednly error message from
        # dictionary
        err_msg = " ".join([
            f"{k}: {' '.join(v)}" for k, v in e.message_dict.items()
        ])
        return HttpResponseBadRequest(
            json.dumps({
                'msg': err_msg
            }),
            content_type="application/json"
        )
    # save folder only after OK validation
    folder.save()
    signals.folder_created.send(
        sender='core.views.documents.create_folder',
        user_id=request.user.id,
        level=logging.INFO,
        message=_("Folder created"),
        folder_id=folder.id
    )

    return HttpResponse(
        json.dumps(
            folder.to_dict()
        )
    )


@json_response
@login_required
@require_POST
def upload(request):
    """
    To understand returned value, have a look at
    papermerge.core.views.decorators.json_reponse decorator
    """
    files = request.FILES.getlist('file')
    if not files:
        logger.warning(
            "POST request.FILES is empty. Forgot adding file?"
        )
        return "Missing input file", 400

    if len(files) > 1:
        msg = "More then one files per ajax? how come?"
        logger.warning(msg)

        return msg, 400

    f = files[0]

    logger.debug("upload for f=%s user=%s", f, request.user)

    user = request.user
    parent_id = request.POST.get('parent', "-1")
    parent_id = filter_node_id(parent_id)

    lang = request.POST.get('language')
    notes = request.POST.get('notes')

    init_kwargs = {'payload': f, 'processor': WEB}

    apply_kwargs = {
        'user': user,
        'name': f.name,
        'parent': parent_id,
        'lang': lang,
        'notes': notes,
        'apply_async': True
    }

    try:
        doc = go_through_pipelines(init_kwargs, apply_kwargs)
    except ValidationError as error:
        return str(error), 400

    if not doc:
        status = 400
        msg = _(
            "File type not supported."
            " Only pdf, tiff, png, jpeg files are supported"
        )
        return msg, status

    # after each upload return a json object with
    # following fields:
    #
    # - title
    # - preview_url
    # - doc_id
    # - action_url  -> needed for renaming/deleting selected item
    #
    # with that info a new thumbnail will be created.
    preview_url = reverse(
        'core:preview', args=(doc.id, 200, 1)
    )

    result = {
        'title': doc.title,
        'doc_id': doc.id,
        'action_url': "",
        'preview_url': preview_url
    }

    return result


@login_required
def usersettings(request, option, value):

    if option == 'documents_view':
        user_settings = request.user.preferences
        if value in ('list', 'grid'):
            user_settings['views__documents_view'] = value
            user_settings['views__documents_view']

    return HttpResponseRedirect(
        request.META.get('HTTP_REFERER')
    )


@login_required
def hocr(request, id, step=None, page="1"):

    logger.debug(f"hocr for doc_id={id}, step={step}, page={page}")
    try:
        doc = Document.objects.get(id=id)
    except Document.DoesNotExist:
        raise Http404("Document does not exists")

    version = request.GET.get('version', None)
    doc_path = doc.path(version=version)

    if request.user.has_perm(Access.PERM_READ, doc):
        # document absolute path
        doc_abs_path = default_storage.abspath(doc_path.url())
        if not os.path.exists(
            doc_abs_path
        ):
            raise Http404("HOCR data not yet ready.")

        page_count = get_pagecount(doc_abs_path)
        if page > page_count or page < 0:
            raise Http404("Page does not exists")

        page_path = doc.page_paths(version=version)[page]
        hocr_abs_path = default_storage.abspath(page_path.hocr_url())

        logger.debug(f"Extract words from {hocr_abs_path}")

        if not os.path.exists(hocr_abs_path):
            default_storage.download(
                page_path.hocr_url()
            )

        if not os.path.exists(hocr_abs_path):
            raise Http404("HOCR data not yet ready.")

        # At this point local HOCR data should be available.
        hocr = Hocr(
            hocr_file_path=hocr_abs_path
        )

        return HttpResponse(
            json.dumps({
                'hocr': hocr.good_json_words(),
                'hocr_meta': hocr.get_meta()
            }),
            content_type="application/json",
        )

    return HttpResponseForbidden()


@login_required
def preview(request, id, step=None, page="1"):

    try:
        doc = Document.objects.get(id=id)
    except Document.DoesNotExist:
        raise Http404("Document does not exists")

    if request.user.has_perm(Access.PERM_READ, doc):
        version = request.GET.get('version', None)

        page_path = doc.get_page_path(
            page_num=page,
            step=Step(step),
            version=version
        )
        img_abs_path = default_storage.abspath(
            page_path.img_url()
        )
        doc_abs_path = default_storage.abspath(
            doc.path().url()
        )

        if not os.path.exists(img_abs_path):
            logger.debug(
                f"Preview image does not exists: {img_abs_path}"
            )
            if not os.path.exists(doc_abs_path):
                logger.debug(f"File does not exists: {doc_abs_path}")
                logger.debug(
                    f"Downloading to {doc_abs_path}."
                )
                default_storage.download(doc.path().url())

            extract_img(
                page_path, media_root=settings.MEDIA_ROOT
            )

        try:
            with open(img_abs_path, "rb") as f:
                return HttpResponse(f.read(), content_type="image/jpeg")
        except IOError:
            generic_file = "admin/img/document.png"
            if Step(step).is_thumbnail:
                generic_file = "admin/img/document_thumbnail.png"

            file_path = finders.find(generic_file)

            with open(file_path, "rb") as f:
                return HttpResponse(f.read(), content_type="image/png")

    return HttpResponseForbidden()


@json_response
@login_required
def text_view(
    request,
    id,
    document_version,
    page_number
):

    try:
        page = Page.objects.get(
            document__id=id,
            number=page_number
        )
    except Page.DoesNotExist:
        raise Http404("Page does not exists")

    doc = page.document

    if request.user.has_perm(Access.PERM_READ, doc):
        txt_abs_path = default_storage.abspath(
            page.path(version=document_version).txt_url()
        )
        text = ""

        with open(txt_abs_path, "r") as f:
            text = f.read()

        return {
            'page_text': text,
            'page_number': page.number,
            'document_version': document_version
        }

    msg = _(
        "%s does not have read perissions on %s"
    ) % (request.user.username, doc.title)

    return msg, HttpResponseForbidden.status_code


@json_response
@login_required
@require_POST
def run_ocr_view(request):

    post_data = json.loads(request.body)
    node_ids = post_data['document_ids']
    new_lang = post_data['lang']

    documents = Document.objects.filter(
        id__in=node_ids
    )
    nodes_perms = request.user.get_perms_dict(
        documents, Access.ALL_PERMS
    )
    for node in documents:
        if not nodes_perms[node.id].get(
            Access.PERM_WRITE, False
        ):
            msg = _(
                "%s does not have write perission on %s"
            ) % (request.user.username, node.title)

            return msg, HttpResponseForbidden.status_code

    for doc in documents:
        old_version = doc.version
        new_version = doc.version + 1

        default_storage.copy_doc(
            src=doc.path(version=old_version),
            dst=doc.path(version=new_version)
        )
        for page_num in range(1, doc.page_count + 1):
            ocr_page.apply_async(kwargs={
                'user_id': doc.user.id,
                'document_id': doc.id,
                'file_name': doc.file_name,
                'page_num': page_num,
                'lang': new_lang,
                'namespace': getattr(default_storage, 'namespace', None),
                'version': new_version
            })

        doc.lang = new_lang
        doc.version = new_version
        doc.save()

    return {'msg': _("OCR process successfully started")}



import inspect
import logging

from django.apps import apps
from django.core import checks
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models.fields.related import (
    ForeignObjectRel,
    OneToOneRel,
    RelatedField
)

from modelcluster.fields import ParentalManyToManyField
from papermerge.search.backends import get_search_backend


logger = logging.getLogger(__name__)


class Indexed:
    @classmethod
    def indexed_get_parent(cls, require_model=True):
        for base in cls.__bases__:
            if issubclass(base, Indexed) and (
                issubclass(base, models.Model) or require_model is False
            ):
                return base

    @classmethod
    def indexed_get_content_type(cls):
        # Work out content type
        content_type = (cls._meta.app_label + '_' + cls.__name__).lower()

        # Get parent content type
        parent = cls.indexed_get_parent()
        if parent:
            parent_content_type = parent.indexed_get_content_type()
            return parent_content_type + '_' + content_type
        else:
            return content_type

    @classmethod
    def indexed_get_toplevel_content_type(cls):
        # Get parent content type
        parent = cls.indexed_get_parent()
        if parent:
            return parent.indexed_get_content_type()
        else:
            # At toplevel, return this content type
            return (cls._meta.app_label + '_' + cls.__name__).lower()

    @classmethod
    def get_search_fields(cls):
        search_fields = {}

        for field in cls.search_fields:
            search_fields[(type(field), field.field_name)] = field

        return list(search_fields.values())

    @classmethod
    def get_searchable_search_fields(cls):
        return [
            field for field in cls.get_search_fields()
            if isinstance(field, SearchField)
        ]

    @classmethod
    def get_autocomplete_search_fields(cls):
        return [
            field for field in cls.get_search_fields()
            if isinstance(field, AutocompleteField)
        ]

    @classmethod
    def get_filterable_search_fields(cls):
        return [
            field for field in cls.get_search_fields()
            if isinstance(field, FilterField)
        ]

    @classmethod
    def get_indexed_objects(cls):
        queryset = cls.objects.all()

        # Add prefetch/select related for RelatedFields
        for field in cls.get_search_fields():
            if isinstance(field, RelatedFields):
                queryset = field.select_on_queryset(queryset)

        return queryset

    def get_indexed_instance(self):
        """
        If the indexed model uses multi table inheritance, override this method
        to return the instance in its most specific class so it reindexes properly.
        """
        return self

    @classmethod
    def _has_field(cls, name):
        try:
            cls._meta.get_field(name)
            return True
        except FieldDoesNotExist:
            return hasattr(cls, name)

    @classmethod
    def check(cls, **kwargs):
        errors = super(Indexed, cls).check(**kwargs)
        errors.extend(cls._check_search_fields(**kwargs))
        return errors

    @classmethod
    def _check_search_fields(cls, **kwargs):
        errors = []
        for field in cls.get_search_fields():
            message = "{model}.search_fields contains non-existent field '{name}'"
            if not cls._has_field(field.field_name):
                errors.append(
                    checks.Warning(
                        message.format(model=cls.__name__, name=field.field_name),
                        obj=cls,
                    )
                )
        return errors

    search_fields = []


def get_indexed_models():
    return [
        model for model in apps.get_models()
        if issubclass(model, Indexed) and not model._meta.abstract
    ]


def class_is_indexed(cls):
    return issubclass(cls, Indexed) and issubclass(cls, models.Model) and not cls._meta.abstract


def get_indexed_instance(instance, check_exists=True):
    indexed_instance = instance.get_indexed_instance()
    if indexed_instance is None:
        return

    # Make sure that the instance is in its class's indexed objects
    if check_exists and not type(indexed_instance).get_indexed_objects().filter(pk=indexed_instance.pk).exists():
        return

    return indexed_instance


def insert_or_update_object(instance):
    indexed_instance = get_indexed_instance(instance)

    if indexed_instance:
        backend = get_search_backend()
        try:
            backend.add(indexed_instance)
        except Exception:
            # Catch and log all errors
            logger.exception(
                "Exception while adding %r into the '%s' search backend",
                indexed_instance,
                backend
            )


def remove_object(instance):
    indexed_instance = get_indexed_instance(instance, check_exists=False)

    if indexed_instance:
        backend = get_search_backend()
        try:
            backend.delete(indexed_instance)
        except Exception:
            # Catch and log all errors
            logger.exception(
                "Exception while deleting %r from the '%s' search backend",
                indexed_instance, backend
            )


class BaseField:
    def __init__(self, field_name, **kwargs):
        self.field_name = field_name
        self.kwargs = kwargs

    def get_field(self, cls):
        return cls._meta.get_field(self.field_name)

    def get_attname(self, cls):
        try:
            field = self.get_field(cls)
            return field.attname
        except FieldDoesNotExist:
            return self.field_name

    def get_definition_model(self, cls):
        try:
            field = self.get_field(cls)
            return field.model
        except FieldDoesNotExist:
            # Find where it was defined by walking the inheritance tree
            for base_cls in inspect.getmro(cls):
                if self.field_name in base_cls.__dict__:
                    return base_cls

    def get_type(self, cls):
        if 'type' in self.kwargs:
            return self.kwargs['type']

        try:
            field = self.get_field(cls)

            # Follow foreign keys to find underlying type
            # We use a while loop as it's possible for a foreign key
            # to target a foreign key in another model.
            # (for example, a foreign key to a child page model will
            # point to the `page_ptr_id` field so we need to follow this
            # second foreign key to find the `id`` field in the Page model)
            while isinstance(field, RelatedField):
                field = field.target_field

            return field.get_internal_type()

        except FieldDoesNotExist:
            return 'CharField'

    def get_value(self, obj):
        from taggit.managers import TaggableManager

        try:
            field = self.get_field(obj.__class__)
            value = field.value_from_object(obj)
            if hasattr(field, 'get_searchable_content'):
                value = field.get_searchable_content(value)
            elif isinstance(field, TaggableManager):
                # As of django-taggit 1.0, value_from_object returns a list of Tag objects,
                # which matches what we want
                pass
            elif isinstance(field, RelatedField):
                # The type of the ForeignKey may have a get_searchable_content method that we should
                # call. Firstly we need to find the field its referencing but it may be referencing
                # another RelatedField (eg an FK to page_ptr_id) so we need to run this in a while
                # loop to find the actual remote field.
                remote_field = field
                while isinstance(remote_field, RelatedField):
                    remote_field = remote_field.target_field

                if hasattr(remote_field, 'get_searchable_content'):
                    value = remote_field.get_searchable_content(value)
            return value
        except FieldDoesNotExist:
            value = getattr(obj, self.field_name, None)
            if hasattr(value, '__call__'):
                value = value()
            return value

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.field_name)


class SearchField(BaseField):
    def __init__(self, field_name, boost=None, partial_match=False, **kwargs):
        super().__init__(field_name, **kwargs)
        self.boost = boost
        self.partial_match = partial_match


class AutocompleteField(BaseField):
    pass


class FilterField(BaseField):
    pass


class RelatedFields:
    def __init__(self, field_name, fields):
        self.field_name = field_name
        self.fields = fields

    def get_field(self, cls):
        return cls._meta.get_field(self.field_name)

    def get_definition_model(self, cls):
        field = self.get_field(cls)
        return field.model

    def get_value(self, obj):
        field = self.get_field(obj.__class__)

        if isinstance(field, (RelatedField, ForeignObjectRel)):
            return getattr(obj, self.field_name)

    def select_on_queryset(self, queryset):
        """
        This method runs either prefetch_related or select_related on the queryset
        to improve indexing speed of the relation.

        It decides which method to call based on the number of related objects:
         - single (eg ForeignKey, OneToOne), it runs select_related
         - multiple (eg ManyToMany, reverse ForeignKey) it runs prefetch_related
        """
        try:
            field = self.get_field(queryset.model)
        except FieldDoesNotExist:
            return queryset

        if isinstance(field, RelatedField) and not isinstance(field, ParentalManyToManyField):
            if field.many_to_one or field.one_to_one:
                queryset = queryset.select_related(self.field_name)
            elif field.one_to_many or field.many_to_many:
                queryset = queryset.prefetch_related(self.field_name)

        elif isinstance(field, ForeignObjectRel):
            # Reverse relation
            if isinstance(field, OneToOneRel):
                # select_related for reverse OneToOneField
                queryset = queryset.select_related(self.field_name)
            else:
                # prefetch_related for anything else (reverse ForeignKey/ManyToManyField)
                queryset = queryset.prefetch_related(self.field_name)

        return queryset



# Generated by Django 3.0.6 on 2020-06-15 05:25

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0014_auto_20200605_1042'),
    ]

    operations = [
        migrations.AddField(
            model_name='kvstore',
            name='kv_format',
            field=models.CharField(blank=True, max_length=64, null=True),
        ),
        migrations.AddField(
            model_name='kvstore',
            name='kv_type',
            field=models.CharField(choices=[('text', 'Text'), ('money', 'Currency'), ('numeric', 'Numeric'), ('date', 'Date')], default='text', max_length=16),
        ),
        migrations.AlterField(
            model_name='page',
            name='document',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='pages', to='core.Document'),
        ),
    ]



import logging
from django.conf import settings
from django.utils.module_loading import import_string

logger = logging.getLogger(__name__)


def get_storage_class(import_path=None):
    return import_string(
        import_path or settings.PAPERMERGE_DEFAULT_FILE_STORAGE
    )


storage_class = get_storage_class()

storage_kwargs = settings.PAPERMERGE_FILE_STORAGE_KWARGS or {}

default_storage = storage_class(
    location=settings.MEDIA_ROOT,
    **storage_kwargs
)



from os.path import getsize, basename
import logging
from magic import from_file
from tempfile import _TemporaryFileWrapper

from django.core.files.temp import NamedTemporaryFile
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import TemporaryUploadedFile
from django.utils import module_loading

from mglib.pdfinfo import get_pagecount
from mglib.exceptions import FileTypeNotSupported

from papermerge.core.models import (
    Folder, Document, User
)
from papermerge.core.storage import default_storage
from papermerge.core.tasks import ocr_page
from papermerge.core import signal_definitions as signals
from papermerge.core.ocr import COMPLETE, STARTED
from papermerge.core.utils import Timer

logger = logging.getLogger(__name__)


REST_API = "REST_API"
WEB = "WEB"
IMAP = "IMAP"
LOCAL = "LOCAL"


class DefaultPipeline:
    """
    Default Pipeline class. It is meant to be extended by apps. Most commonly
    the methods check_mimetype (remember to change get_pagecount as well) and
    apply should be modified. All checks whether the payload is compatible
    with an extended pipeline is to be done  in the init method, the apply
    method is not expected to raise exceptions.
    """

    def __init__(
        self,
        payload=None,
        doc=None,
        processor=WEB,
        **kwargs
    ):
        """
        Init method of the pipeline. Only succeeds if the file is compatible
        with the pipeline.

        Args:
            payload (Union[bytes, TemporaryUploadedFile,
            _TemporaryFileWrapper], optional): payload to be ingested.
            Defaults to None. doc (Document, optional): document to be
            updated. Defaults to None. processor (str, optional): from which
            importer this class is invocated. Defaults to WEB.

        Raises:
            TypeError: raised when payload is not a supported file object
            FileTypeNotSupported: raised when payload is of a wrong mimetype
            for the pipeline
        """
        self.processor = processor
        self.doc = doc

        if payload is None:
            raise TypeError
        elif isinstance(payload, bytes):
            payload = self.write_temp(payload)

        self.payload = payload

        if isinstance(payload, TemporaryUploadedFile):
            self.path = payload.temporary_file_path()
        elif isinstance(payload, _TemporaryFileWrapper):
            self.path = payload.name
        else:
            raise TypeError

        self.check_mimetype()

    def check_mimetype(self):
        """Check if mimetype of the document to be imported is supported
        by Papermerge or one of its apps.

        Raises:
            FileTypeNotSupported: If the mimetype is not supported by this
            pipeline

        """
        supported_mimetypes = settings.PAPERMERGE_MIMETYPES
        mime = from_file(self.path, mime=True)
        if mime in supported_mimetypes:
            return None
        raise FileTypeNotSupported

    def write_temp(self, payload):
        """Write a temporary file to disk, necessary for certain
        payload types that are not stored on file.

        Args:
            payload (bytes): ingested payload

        Returns:
            temp (NamedTemporaryFile): temporary file on disk
        """
        logger.debug(
            f"{self.processor} importer: creating temporary file"
        )

        temp = NamedTemporaryFile()
        temp.write(payload)
        temp.flush()
        return temp

    @staticmethod
    def get_user_properties(user):
        """Get properties of the document owner, if no owner is specified
        the document gets assigned to first superuser

        Args:
            user (User): owner object

        Returns:
            user (User): owner object
            lang (str): user language
            inbox (Folder): inbox folder
        """
        if user is None:
            user = User.objects.filter(
                is_superuser=True
            ).first()
        if isinstance(user, str):
            user = User.objects.filter(
                username=user
            ).first()
        lang = user.preferences['ocr__OCR_Language']

        inbox, _ = Folder.objects.get_or_create(
            title=Folder.INBOX_NAME,
            parent=None,
            user=user
        )
        return user, lang, inbox

    def move_tempfile(self, doc):
        default_storage.copy_doc(
            src=self.path,
            dst=doc.path().url()
        )
        return None

    def page_count(self):
        return get_pagecount(self.path)

    def ocr_document(
        self,
        document,
        page_count,
        lang
    ):
        user_id = document.user.id
        document_id = document.id
        file_name = document.file_name

        logger.debug(
            f"{self.processor} importer: "
            f"document {document_id} has {page_count} pages."
        )
        for page_num in range(1, page_count + 1):
            signals.page_ocr.send(
                sender='worker',
                level=logging.INFO,
                message="",
                user_id=user_id,
                document_id=document_id,
                page_num=page_num,
                lang=lang,
                status=STARTED
            )

            with Timer() as time:
                ocr_page(
                    user_id=user_id,
                    document_id=document_id,
                    file_name=file_name,
                    page_num=page_num,
                    lang=lang,
                )

            msg = "{} importer: OCR took {} seconds to complete.".format(
                self.processor,
                time
            )
            signals.page_ocr.send(
                sender='worker',
                level=logging.INFO,
                message=msg,
                user_id=user_id,
                document_id=document_id,
                page_num=page_num,
                lang=lang,
                status=COMPLETE
            )

    def get_init_kwargs(self):
        """Propagates keyword arguments to use in the init method
        of donwstream pipelines. Should be overwritten by inheriting
        classes.

        Returns:
            A dict with the generated document or None if no document
            could be generated
        """
        if self.doc:
            return {'doc': self.doc}
        return None

    def get_apply_kwargs(self):
        """Propagates keyword arguments to use in the apply method
        of donwstream pipelines. Should be overwritten by inheriting
        classes.

        Returns:
            None
        """
        return None

    def apply(
        self,
        user=None,
        parent=None,
        lang=None,
        notes=None,
        name=None,
        skip_ocr=False,
        apply_async=False,
        create_document=True,
        **kwargs
    ):
        """
        Apply the pipeline. The document is created or modified here.  This
method is not supposed to throw errors.

        Arguments:
        - user (User, optional): document owner.
        - parent (Folder, optional): folder containing the document.
        - lang (str, optional): OCR language.
        - notes (str, optional): document notes.
        - name (str, optional): document name.
        - skip_ocr (bool, optional):
            whether to skip OCR processing. Defaults to False.
        - apply_async (bool, optional):
            whether to apply OCR asynchronously.
            Defaults to False.
        - create_document (bool, optional): whether to
        create or update a document. Defaults to True.

        Returns:
            Document: the created or updated document
        """
        if parent is None:
            user, lang, inbox = self.get_user_properties(user)
            # in case of upload via REST API, LOCAL, or IMAP interface,
            # documents must land in user's inbox
            if self.processor in (REST_API, LOCAL, IMAP):
                parent = inbox.id
        if name is None:
            name = basename(self.path)
        page_count = self.page_count()
        size = getsize(self.path)

        if create_document and self.doc is None:
            try:
                doc = Document.objects.create_document(
                    user=user,
                    title=name,
                    size=size,
                    lang=lang,
                    file_name=name,
                    parent_id=parent,
                    page_count=page_count,
                    notes=notes
                )
                self.doc = doc
            except ValidationError as error:
                logger.error(f"{self.processor} importer: validation failed")
                raise error
        elif self.doc is not None:
            doc = self.doc
            doc.version = doc.version + 1
            doc.page_count = page_count
            doc.file_name = name
            doc.size = size
            doc.save()
            try:
                doc.recreate_pages()
            except ValueError:
                doc.create_pages()
            doc.full_clean()

        self.move_tempfile(doc)
        self.payload.close()
        if not skip_ocr:

            namespace = default_storage.upload(
                doc_path_url=doc.path().url()
            )

            if apply_async:
                for page_num in range(1, page_count + 1):
                    ocr_page.apply_async(kwargs={
                        'user_id': user.id,
                        'document_id': doc.id,
                        'file_name': name,
                        'page_num': page_num,
                        'lang': lang,
                        'namespace': namespace
                    })
            else:
                self.ocr_document(
                    document=doc,
                    page_count=page_count,
                    lang=lang,
                )

        logger.debug(f"{self.processor} importer: import complete.")
        return doc


def go_through_pipelines(init_kwargs, apply_kwargs):
    """
    Method to go through all the loaded pipelines **in order**. The init and
    apply dictionaries are not reset for each pipeline, they are updated with
    the results of get_init_kwargs and get_apply_kwargs. This means arguments
    need to be set to None by pipelines as well.

    Args:
        init_kwargs (dict): initial init_kwargs
        apply_kwargs (dict): initial apply_kwargs

    Returns:
        Document: created document, needs to be unique for each payload
    """
    processor = init_kwargs.get('processor', WEB)
    doc = None
    pipelines = settings.PAPERMERGE_PIPELINES
    logger.info(f"{processor} importer: importing file")

    for pipeline in pipelines:

        try:
            pipeline_class = module_loading.import_string(pipeline)
        except ImportError:
            logger.error(
                f"{pipeline} could not be loaded."
                " Check if it is installed properly."
            )
            continue
        try:
            importer = pipeline_class(**init_kwargs)
        except TypeError:
            logger.debug(f"{processor} importer: not a file")
            break
        except FileTypeNotSupported:
            logger.debug(f"{processor} importer: filetype not supported")
            continue

        doc = importer.apply(**apply_kwargs)
        logger.info(f"{processor} importer: payload processed successfully")

        init_kwargs_temp = importer.get_init_kwargs()
        apply_kwargs_temp = importer.get_apply_kwargs()

        if init_kwargs_temp:
            init_kwargs = {**init_kwargs, **init_kwargs_temp}
        if apply_kwargs_temp:
            apply_kwargs = {**apply_kwargs, **apply_kwargs_temp}

    return doc



from django.urls import reverse
from papermerge.core import __version__ as PAPERMERGE_VERSION
from papermerge.core.models import (
    Folder,
    Tag
)
from .forms import AdvancedSearchForm
from .registries import (
    user_menu_registry,
    navigation
)

ROLE_DEPENDENT_SIDEBAR_MENU = [
    {
        'required_perm': 'core.view_automate',
        'url': reverse('admin:automates'),
        'title': 'Automates',
        'css_icon_class': 'fa-robot',
        'activate_on': 'automates, automate-add, automate-update'
    },
    {
        'required_perm': None,
        'url': reverse('admin:tags'),
        'title': 'Tags',
        'css_icon_class': 'fa-tag',
        'activate_on': 'tags, tag-add, tag-update'
    },
    {
        'required_perm': None,
        'url': reverse('admin:logs'),
        'title': 'Logs',
        'css_icon_class': 'fa-lightbulb',
        'activate_on': 'logs, log-update'
    },
    {
        'required_perm': 'auth.view_group',
        'url': reverse('admin:groups'),
        'title': 'Groups',
        'css_icon_class': 'fa-users',
        'activate_on': 'groups, group-add, group-update'
    },
    {
        'required_perm': 'core.view_user',
        'url': reverse('admin:users'),
        'title': 'Users',
        'css_icon_class': 'fa-user-friends',
        'activate_on': 'users, user-add, user-update'
    },
    {
        'required_perm': 'core.view_role',
        'url': reverse('admin:roles'),
        'title': 'Roles',
        'css_icon_class': 'fa-user',
        'activate_on': 'roles, role-add, role-update'
    },
]


def sidebar_menu(request):
    """
    Builds sidebar menu depending on user role/permissions

    Sidebar menu is applicable only for authenticated users.
    """
    menu = []

    # menu item has following keys

    # 1. url
    # 2. title
    # 3. css_icon_class
    # 4. activate_on
    if request.user.is_anonymous:
        # main sidebar menu is applicable only for authenticated users
        return {}

    for menu_item in ROLE_DEPENDENT_SIDEBAR_MENU:
        req_perm = menu_item['required_perm']

        if not req_perm:
            menu.append(menu_item)

        if request.user.has_perm(req_perm):
            menu.append(menu_item)

    """
    lte_menu option determines the state of left menu of admin lte3
    (collapsed/expanded). It is used to decide which css class to set
    on <body> element (in admin/base.html).
    Why? Because with client side only solution, the user can see
    for fraction of second other state
    (during refresh -> collapse state -> ~ 0.2 sec -> expanded state)
    """

    return {
        'sidebar_menu': menu,
        'lte_menu': request.COOKIES.get('lte-menu', None)
    }


def user_menu(request):
    values = list(user_menu_registry.values())

    return {
        'user_menu': values
    }


def leftside_navigation(request):

    values = list(navigation.values())

    return {
        'leftside_navigation': values
    }


def extras(request):

    if request.user.is_anonymous:
        return {
            'inbox_count': -1
        }

    try:
        inbox = Folder.objects.get(
            title=Folder.INBOX_NAME,
            user=request.user
        )
        count = inbox.get_children().count()
    except Folder.DoesNotExist:
        count = -1

    pinned_tags = Tag.objects.filter(
        pinned=True,
        user=request.user
    )
    alltags = Tag.objects.filter(
        user=request.user
    )
    form = AdvancedSearchForm(user=request.user)

    return {
        'inbox_count': count,
        'pinned_tags': pinned_tags,
        'papermerge_version': PAPERMERGE_VERSION,
        'search_form': form,
        'alltags': alltags
    }


def user_perms(request):
    if request.user.is_anonymous:
        return {
            'has_perm_change_user': False,
            'has_perm_auth_authtoken': False
        }

    change_user = request.user.has_perm(
        'core.change_user',
    )
    auth_token = request.user.has_perm(
        'knox.view_authtoken',
    )
    # if user has permissions to view/use access
    # featues
    access_feature = request.user.has_perm(
        'core.view_access'
    )
    new_folder = request.user.has_perm(
        'core.add_folder'
    )
    return {
        'has_perm_change_user': change_user,
        'has_perm_view_authtoken': auth_token,
        'has_perm_access_feature': access_feature,
        'has_perm_new_folder': new_folder
    }



from django.contrib import auth
from django.core.exceptions import PermissionDenied
from django.contrib.auth.models import AbstractUser
from django.db import models
from django.utils.translation import gettext_lazy as _

from allauth.account.models import EmailAddress

from papermerge.core.models.automate import Automate
from papermerge.core.models.access import Access
from papermerge.core.models.diff import Diff
from papermerge.core.models.document import Document, AbstractDocument
from papermerge.core.models.folder import Folder
from papermerge.core.models.kvstore import KV, KVPage, KVStoreNode, KVStorePage
from papermerge.core.models.node import BaseTreeNode, AbstractNode
from papermerge.core.models.page import Page
from papermerge.core.models.tags import (
    ColoredTag,
    Tag
)
from papermerge.core.models.role import Role
from papermerge.core.models.sidebar_part import SidebarPart


# A few helper functions for common logic between User and AnonymousUser.
def _user_get_permissions(user, obj, from_name):
    permissions = set()
    name = 'get_%s_permissions' % from_name
    for backend in auth.get_backends():
        if hasattr(backend, name):
            permissions.update(getattr(backend, name)(user, obj))
    return permissions


def _user_has_perm(user, perm, obj):
    """
    A backend can raise `PermissionDenied` to short-circuit permission
    checking.
    """
    for backend in auth.get_backends():
        if not hasattr(backend, 'has_perm'):
            continue
        try:
            if backend.has_perm(user, perm, obj):
                return True
        except PermissionDenied:
            return False
    return False


def _get_perms_dict(user, perms, obj_list):
    #
    #  Bulk permissions return. Optimization measure
    #  for case when folder contains many files.
    #
    for backend in auth.get_backends():
        if not hasattr(backend, 'get_perms_dict'):
            continue

        return backend.get_perms_dict(user, perms, obj_list)


def _user_has_module_perms(user, app_label):
    """
    A backend can raise `PermissionDenied` to short-circuit permission
    checking.
    """
    for backend in auth.get_backends():
        if not hasattr(backend, 'has_module_perms'):
            continue
        try:
            if backend.has_module_perms(user, app_label):
                return True
        except PermissionDenied:
            return False
    return False


class User(AbstractUser):
    # Role is optional.
    # All users EXCEPT superuser
    # have associated a role
    role = models.ForeignKey(
        'Role',
        verbose_name='role',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text=_(
            "Determines what operations this user is"
            " authorized to perform."
        )
    )

    created_at = models.DateTimeField(
        auto_now_add=True
    )
    updated_at = models.DateTimeField(
        auto_now=True
    )

    def update_current_storage(self):
        user_docs = Document.objects.filter(user=self)
        self.current_storage_size = sum(int(doc.size) for doc in user_docs)
        self.save()

    def get_user_permissions(self, obj=None):
        """
        Return a list of permission strings that this user has directly.
        Query all available auth backends. If an object is passed in,
        return only permissions matching this object.
        """
        return _user_get_permissions(self, obj, 'user')

    def get_group_permissions(self, obj=None):
        """
        Return a list of permission strings that this user has through their
        groups. Query all available auth backends. If an object is passed in,
        return only permissions matching this object.
        """
        return _user_get_permissions(self, obj, 'group')

    def get_all_permissions(self, obj=None):
        return _user_get_permissions(self, obj, 'all')

    def has_perm(self, perm, obj=None):
        """
        Return True if the user has the specified permission. Query all
        available auth backends, but return immediately if any backend returns
        True. Thus, a user who has permission from a single auth backend is
        assumed to have permission in general. If an object is provided, check
        permissions for that object.
        """
        return _user_has_perm(self, perm, obj)

    def get_perms_dict(self, obj, perms):
        return _get_perms_dict(self, obj, perms)

    def has_perms(self, perm_list, obj=None):
        """
        Return True if the user has each of the specified permissions. If
        object is passed, check if the user has all required perms for it.
        """
        return all(self.has_perm(perm, obj) for perm in perm_list)

    def has_module_perms(self, app_label):
        """
        Return True if the user has any permissions in the given app label.
        Use similar logic as has_perm(), above.
        """
        # Active superusers have all permissions.
        if self.is_active and self.is_superuser:
            return True

        return _user_has_module_perms(self, app_label)

    def confirm_email(self):
        """
        Allauth requires an associated EmailAddress object with
        ``verified`` attribute set to ``true`` in order
        to consider that user confirmed its email.
        This method is required in case email confirmation is
        enabled because allauth cannot distinguish between
        users created by administrative person from
        users created by web form registration. In both cases
        it will require email confirmation.
        """
        email_address, _ = EmailAddress.objects.get_or_create(
            email=self.email, user=self
        )
        email_address.verified = True
        email_address.user = self
        email_address.save()


__all__ = [
    'User',
    'Role',
    'Folder',
    'Document',
    'AbstractDocument',
    'Page',
    'AbstractNode',
    'BaseTreeNode',
    'Access',
    'Automate',
    'Diff',
    'KV',
    'KVPage',
    'KVStoreNode',
    'KVStorePage',
    'ColoredTag',
    'Tag',
    'SidebarPart',
]



# Generated by Django 3.0.10 on 2020-09-21 04:56

from django.db import migrations, models
import taggit.managers


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0022_auto_20200909_0826'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='tag',
            options={'ordering': ['name'], 'verbose_name': 'Tag', 'verbose_name_plural': 'Tags'},
        ),
        migrations.RemoveField(
            model_name='automate',
            name='plugin_name',
        ),
        migrations.AddField(
            model_name='automate',
            name='tags',
            field=taggit.managers.TaggableManager(help_text='A comma-separated list of tags.', through='core.ColoredTag', to='core.Tag', verbose_name='Tags'),
        ),
        migrations.AlterField(
            model_name='tag',
            name='pinned',
            field=models.BooleanField(default=False, help_text='Pinned tag will be displayed under Documents menu. It serves as shortcut to quickly filter folders/documents associated with this tag'),
        ),
    ]



# Generated by Django 3.0.8 on 2020-08-28 11:48

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='LogEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('message', models.TextField()),
                ('level', models.PositiveIntegerField(choices=[(10, 'Debug'), (20, 'Info'), (30, 'Warning'), (40, 'Error'), (50, 'Critical')], default=20)),
                ('action_time', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='action time')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'log entry',
                'verbose_name_plural': 'log entries',
                'ordering': ('-action_time',),
            },
        ),
    ]



import logging

from django.dispatch import receiver
from django.utils.translation import gettext as _

from papermerge.core.models import Folder

from papermerge.core.signal_definitions import (
    folder_created,
    nodes_deleted,
)
from papermerge.contrib.admin.models import LogEntry

logger = logging.getLogger(__name__)


@receiver(folder_created)
def folder_created_handler(sender, **kwargs):
    folder_id = kwargs.get('folder_id')
    user_id = kwargs.get('user_id')
    level = kwargs.get('level')

    folder = Folder.objects.get(id=folder_id)

    folder_title = folder.title

    msg = _(
        "Node/Folder %(folder_title)s created. Folder id=%(folder_id)s."
    ) % {
        'folder_title': folder_title,
        'folder_id': folder_id
    }

    LogEntry.objects.create(
        user_id=user_id,
        level=level,
        message=msg
    )


@receiver(nodes_deleted)
def nodes_deleted_handler(sender, **kwargs):
    node_titles = kwargs.get('node_titles')
    user_id = kwargs.get('user_id')
    level = kwargs.get('level')
    node_ids = kwargs.get('node_ids')
    msg = _(
        "Node(s) %(node_titles)s were deleted. Node ids=%(node_ids)s"
    ) % {
        'node_titles': ', '.join(node_titles),
        'node_ids': node_ids
    }

    LogEntry.objects.create(
        user_id=user_id,
        level=level,
        message=msg
    )



from pytz import common_timezones

from django.conf import settings

from dynamic_preferences.preferences import Section as OrigSection
from dynamic_preferences.types import (
    ChoicePreference,
    BooleanPreference,
    StringPreference
)
from dynamic_preferences.users.registries import user_preferences_registry

from .lib.lang import get_ocr_lang_choices, get_default_ocr_lang


def _get_timezone_choices():
    return list((tz, tz) for tz in common_timezones)


def _is_email_routing_enabled():

    by_user = getattr(
        settings,
        'PAPERMERGE_IMPORT_MAIL_BY_USER',
        False
    )

    by_secret = getattr(
        settings,
        'PAPERMERGE_IMPORT_MAIL_BY_SECRET',
        False
    )
    return by_user or by_secret


class Section(OrigSection):
    """
    Section has following attributes:

        * name
        * verbose_name
        * help_text
        * icon_name
        * visible (default=True)

    If Section has visibility attribute set to False - it won't be
    displayed/visible in User preferences view.
    Example:
    if administrator decides to disable email_routing feature (i.e. he/she
    preferes all incoming email to land in superuser Inbox) with:

        PAPERMERGE_IMPORT_MAIL_BY_USER = False
        PAPERMERGE_IMPORT_MAIL_BY_SECRET = False

    Then, there is no point (it is even confusing) for user to fiddle with
    Email Routing options. From UX poing of view - it is better not to display
    Email Routing section at all.
    """

    def __init__(
        self,
        name,
        verbose_name=None,
        help_text=None,
        icon_name=None,
        visible=True,
    ):
        super().__init__(
            name=name,
            verbose_name=verbose_name
        )
        self.help_text = help_text
        self.icon_name = icon_name
        self.visible = visible


localization = Section(
    'localization',
    verbose_name="Localization",
    icon_name="globe-americas",
    help_text="Timezone, date and time formats"
)

ocr = Section(
    'ocr',
    verbose_name="Opical Character Recognition",
    icon_name="eye",
    help_text="Choose default OCR Language"
)

email_routing = Section(
    'email_routing',
    verbose_name="Email Routing",
    icon_name="envelope-open-text",
    help_text="How email attachments match your Inbox",
    visible=_is_email_routing_enabled()
)


@user_preferences_registry.register
class TimezoneGlobal(ChoicePreference):
    help_text = """
    Timezone
"""
    section = localization
    name = "timezone"
    choices = _get_timezone_choices()
    # fallback value
    default = settings.TIME_ZONE


@user_preferences_registry.register
class OcrLanguage(ChoicePreference):
    help_text = """
    Language used for OCR processing.
"""
    section = ocr
    name = 'OCR_Language'
    choices = get_ocr_lang_choices()
    default = get_default_ocr_lang()


@user_preferences_registry.register
class LocalizationDate(ChoicePreference):
    help_text = """
    Date format
"""
    section = localization
    name = 'date_format'
    choices = (
        ('%Y-%m-%d', '2020-11-25'),
        ('%a %d %b, %Y', 'Wed 25 Nov, 2020'),
        ('%d %b, %Y', '25 Nov, 2020'),
        ('%m/%d/%Y', '11/25/2020'),
        ('%d/%m/%Y', '25/11/2020'),
        ('%d.%m.%y', '25.11.20'),
        ('%d.%m.%Y', '25.11.2020'),
    )
    default = '%Y-%m-%d'


@user_preferences_registry.register
class LocalizationTime(ChoicePreference):
    help_text = """
    Time format
"""
    section = localization
    name = 'time_format'
    choices = (
        ('%I:%M %p', '9:48 PM'),
        ('%H:%M', '21:48'),
    )
    default = '%H:%M'


@user_preferences_registry.register
class EmailRoutingByUser(BooleanPreference):
    help_text = """
    Email attachments will end up in your Inbox
    if incoming email's 'From' or 'To' fields matches your Papermerge user
    email address.
"""
    section = email_routing
    name = "by_user"
    default = False


@user_preferences_registry.register
class EmailRoutingBySecret(BooleanPreference):
    help_text = """
    Email attachments will end up in your will Inbox
    if given secret text is found either in email body or
    in email subject field.
"""
    section = email_routing
    name = "by_secret"
    default = False


@user_preferences_registry.register
class StringPreference(StringPreference):
    help_text = """
    Email secret text used by 'routing by secret' option.
"""
    section = email_routing
    name = "mail_secret"
    default = ""





class SidebarPart:

    """
    Wrapper class for managing/rendering document parts
    on sidebar.
    """

    def __init__(self, document):
        # papermerge.core.models.document instance
        self.document = document

    def to_json(self):
        pass



from django.http import Http404
from django.utils.translation import gettext as _

from rest_framework.views import APIView
from rest_framework import status
from rest_framework.response import Response
from rest_framework.parsers import (
    JSONParser,
    MultiPartParser,
    FileUploadParser
)
from rest_framework.permissions import IsAuthenticated

from papermerge.core.models import (
    Document, Access
)
from papermerge.core.serializers import DocumentSerializer
from papermerge.core.import_pipeline import (
    go_through_pipelines,
    REST_API
)


class PagesView(APIView):
    permission_classes = [IsAuthenticated]

    def delete(self, request, doc_id):
        """
        Deletes Pages from doc_id document
        """
        try:
            doc = Document.objects.get(id=doc_id)
        except Document.DoesNotExist:
            raise Http404("Document does not exists")

        if request.user.has_perm(
            Access.PERM_WRITE, doc
        ):
            page_nums = request.GET.getlist('pages[]')
            page_nums = [int(number) for number in page_nums]

            doc.delete_pages(page_numbers=page_nums)

            return Response(status=status.HTTP_204_NO_CONTENT)

        return Response(status=status.HTTP_401_UNAUTHORIZED)

    def post(self, request, doc_id):
        """
        Reorders pages from doc_id document

        request.data is expected to be a list of dictionaries:
        Example:
            [
                {page_num: 2, page_order: 1},
                {page_num: 1, page_order: 2},
                {page_num: 3, page_order: 3}
            ]
        which reads:
            page number 2, must become page number 1
            page number 1, must become page number 2
            page number 3 stays same.
        """
        try:
            doc = Document.objects.get(id=doc_id)
        except Document.DoesNotExist:
            raise Http404("Document does not exists")

        if request.user.has_perm(
            Access.PERM_WRITE, doc
        ):
            doc.reorder_pages(request.data)

            return Response(status=status.HTTP_204_NO_CONTENT)

        return Response(status=status.HTTP_403_FORBIDDEN)


class PagesCutView(APIView):
    permission_classes = [IsAuthenticated]

    def post(self, request, doc_id):
        """
        Cut Pages from doc_id document.
        It adds pages to clipboard designated for pages
        which where cut.
        """
        try:
            doc = Document.objects.get(id=doc_id)
        except Document.DoesNotExist:
            raise Http404("Document does not exists")

        if request.user.has_perm(
            Access.PERM_WRITE, doc
        ):
            page_nums = request.data
            page_nums = [int(number) for number in page_nums]

            # request.clipboard.pages.add(...)
            request.pages.add(doc_id=doc_id, page_nums=page_nums)

            return Response(status=status.HTTP_204_NO_CONTENT)

        return Response(
            status=status.HTTP_403_FORBIDDEN,
            data={
                'msg': _(
                    "You don't have permissions to cut pages"
                    " in this document."
                )
            }
        )


class PagesPasteView(APIView):
    permission_classes = [IsAuthenticated]

    def post(self, request, doc_id):
        """
        Paste pages (within document view).
        NO new document is created.
        Code for pasting document in changelist view (i.e. when
        a NEW document is created) is in
        papermerge.core.views.documents.paste_pages
        """
        before = request.POST.get('before', False)
        after = request.POST.get('after', False)

        try:
            document = Document.objects.get(id=doc_id)
        except Document.DoesNotExist:
            raise Http404("Document does not exists")

        if request.user.has_perm(
            Access.PERM_WRITE, document
        ):
            Document.paste_pages(
                user=request.user,
                parent_id=document.parent,
                dst_document=document,
                doc_pages=request.pages.all(),
                before=before,
                after=after
            )

            request.pages.clear()

            return Response(status=status.HTTP_204_NO_CONTENT)

        return Response(
            status=status.HTTP_403_FORBIDDEN,
            data={
                'msg': _(
                    "You don't have permissions to paste pages"
                    " in this document."
                )
            }
        )


class DocumentsView(APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request):
        docs_list = []
        docs = Document.objects.all()

        docs_perms = request.user.get_perms_dict(
            docs, Access.ALL_PERMS
        )

        for doc in docs:
            if docs_perms[doc.id].get(Access.PERM_READ, False):
                doc_dict = doc.to_dict()
                docs_list.append(
                    doc_dict
                )

        return Response(docs_list)


class DocumentUploadView(APIView):
    """
    REST API for uploading a file.
    """
    permission_classes = [IsAuthenticated]
    parser_classes = [FileUploadParser, MultiPartParser]

    def put(self, request, filename):

        file_obj = request.data['file']
        init_kwargs = {'payload': file_obj, 'processor': REST_API}

        apply_kwargs = {
            'user': request.user.username,
            'name': filename,
            'apply_async': True
        }
        doc = go_through_pipelines(
            init_kwargs=init_kwargs,
            apply_kwargs=apply_kwargs
        )

        if isinstance(doc, Document):
            serializer = DocumentSerializer(doc)
            return Response(serializer.data)

        return Response(status=200)


class DocumentView(APIView):

    permission_classes = [IsAuthenticated]

    def get_object(self, pk):
        try:
            return Document.objects.get(pk=pk)
        except Document.DoesNotExist:
            raise Http404

    def get(self, request, pk):
        """
        Retrieve, update or delete a document.
        """
        document = self.get_object(pk)

        if request.user.has_perm(Access.PERM_READ, document):
            serializer = DocumentSerializer(document)
            return Response(serializer.data)

        return Response(
            status=status.HTTP_403_FORBIDDEN
        )

    def put(self, request, pk):
        data = JSONParser().parse(request)
        document = self.get_object(pk)

        if request.user.has_perm(Access.PERM_WRITE, document):
            serializer = DocumentSerializer(document, data=data)

            if serializer.is_valid():
                serializer.save()
                return Response(serializer.data)

        return Response(
            status=status.HTTP_403_FORBIDDEN
        )

    def delete(self, request, pk):

        document = self.get_object(pk)

        if request.user.has_perm(Access.PERM_DELETE, document):
            document.delete()
            return Response(status=status.HTTP_200_OK)

        return Response(
            status=status.HTTP_403_FORBIDDEN
        )



import logging
import re

from django.db import models
from django.utils.translation import ugettext_lazy as _

from .document import Document
from .folder import Folder
from .tags import ColoredTag, UserTaggableManager
from .access import Access
from papermerge.core.signal_definitions import automates_matching


logger = logging.getLogger(__name__)


class AutomateQuerySet(models.QuerySet):

    def run(self, pages_qs):
        """
        Runs (scoped by queryset) automates
        over given page models.

        This method is invoked when user
        wants to re-run (eventually modified) selected
        automates over documents (well, pages) from its
        inbox.
        """
        total_matched = []

        for page in pages_qs:
            matched = []
            document = page.document
            page_num = page.number
            text = page.text

            for automate in self.all():
                if automate.is_a_match(text):
                    automate.apply(
                        document, page_num, text
                    )
                    matched.append(automate)
                    total_matched.append(automate)

            message = ""
            automates = Automate.objects.filter(
                pk__in=[item.pk for item in matched]
            )

            message = _(
                "%(count)s of %(total)s Automate(s) matched. ") % {
                'count': len(matched),
                'total': automates.count()
            }

            if len(matched) > 0:
                message += _(
                    "List of matched Automates: %(matched_automates)s"
                ) % {
                    'matched_automates': matched
                }

            automates_matching.send(
                sender="papermerge.core.automate",
                user_id=document.user.id,
                document_id=document.id,
                level=logging.INFO,
                message=message,
                page_num=page_num,
                text=text
            )

        return Automate.objects.filter(
            pk__in=[item.pk for item in total_matched]
        )


class AutomateManager(models.Manager):

    def get_queryset(self):
        return AutomateQuerySet(self.model, using=self._db)


class Automate(models.Model):

    """
    Automates:
        * moving of documents into destination folder
        * extraction of metadata
    """

    # any match - looks for any occurrence of any word
    # provided in the document
    MATCH_ANY = 1
    # all match - looks for all occurrences of all words
    # provided in the document (order does not matter)
    MATCH_ALL = 2
    # literal match means that the text you enter must appear in
    # the document exactly as you've entered it
    MATCH_LITERAL = 3
    # reg exp match
    MATCH_REGEX = 4

    MATCHING_ALGORITHMS = (
        (MATCH_ANY, _("Any")),
        (MATCH_ALL, _("All")),
        (MATCH_LITERAL, _("Literal")),
        (MATCH_REGEX, _("Regular Expression")),
    )

    name = models.CharField(
        _('Name'),
        max_length=128,
        unique=True
    )

    # text to match
    match = models.CharField(
        _('Match'),
        max_length=256,
        blank=True
    )

    matching_algorithm = models.PositiveIntegerField(
        _('Matching Algorithm'),
        choices=MATCHING_ALGORITHMS,
        default=MATCH_ANY,
    )

    # shoud be matching case_sensitive? i.e. uppercase == lowercase
    is_case_sensitive = models.BooleanField(
        _('Is case sensitive'),
        default=True
    )

    tags = UserTaggableManager(
        through=ColoredTag,
        blank=True  # tags are optional
    )

    dst_folder = models.ForeignKey(
        'Folder',
        on_delete=models.DO_NOTHING,
        verbose_name=_('Destination Folder'),
        blank=True,  # destination folder is optional as well
        null=True
    )

    # Should this page be cutted and pasted as separate document?
    # Very useful in case of bulk receipts scans
    extract_page = models.BooleanField(
        _('Extract Page'),
        default=False
    )

    user = models.ForeignKey(
        'User',
        models.CASCADE,
        blank=True,
        null=True
    )

    objects = AutomateManager()

    def __str__(self):
        return self.name

    def is_a_match(self, hocr):
        # Check that match is not empty
        if self.match.strip() == "":
            return False

        search_kwargs = {}
        if not self.is_case_sensitive:
            search_kwargs = {"flags": re.IGNORECASE}

        if self.matching_algorithm == Automate.MATCH_ANY:
            return self._match_any(hocr, search_kwargs)

        if self.matching_algorithm == Automate.MATCH_ALL:
            return self._match_all(hocr, search_kwargs)

        if self.matching_algorithm == Automate.MATCH_LITERAL:
            return self._match_literal(hocr, search_kwargs)

        if self.matching_algorithm == Automate.MATCH_REGEX:
            return self._match_regexp(hocr, search_kwargs)

        return False

    def move_to(self, document, dst_folder):
        document.refresh_from_db()
        dst_folder.refresh_from_db()

        if document and dst_folder:
            Document.objects.move_node(
                document, dst_folder
            )

    def is_automate_applicable(self, document):
        """
        Automation is only applicable:

        1. for documents which are in document's user inbox folder
        2. for documents  which are in one of the folders
        mentioned in automates.dst_folder.

        If 2. is ignored unexpected things might happed: consider
        this case:

            -> a. 5 page document D matches automate criteria on all pages.
            -> b. metadata will be found only on the first page
            -> c. first page is OCRed last

        Let's assume that second page is processed first and it matches
        automation criteria.
        This means that document D will be moved to destination folder.
        If only 1. is considered, all other pages of document D will not
        be applicable for automation!

        What about making all documents applicable for automation?
        In this last case, if user uploads a document to a folder M - documents
        may be moved unexpectedly by background automation process! From user
        point of view - docs will just vanish!
        """
        if not document.parent:
            return False

        if document.parent.title == Folder.INBOX_NAME:
            # document is in Inbox
            logger.debug(f"Document {document} is in Inbox")
            return True

        for automate in Automate.objects.filter(user=document.user):
            # document is in one of automation dst_folders
            if document.parent == automate.dst_folder:
                logger.debug(f"Document {document} is in one of dst folders")
                return True

        return False

    def apply(
        self,
        document,
        page_num,
        text
    ):
        logger.debug("automate.Apply begin")

        if not self.is_automate_applicable(document):
            logger.debug("Automate not applicable. Quit.")
            return

        if self.dst_folder:
            if not self.user.has_perm(Access.PERM_WRITE, self.dst_folder):
                logger.debug("User does not have write access")
                return

            self.move_to(
                document,
                self.dst_folder
            )

        _tags = [tag.name for tag in self.tags.all()]
        document.tags.add(
            *_tags,
            tag_kwargs={'user': self.user}
        )

    def _match_any(self, hocr, search_kwargs):
        for word in self._split_match():
            regexp = r"\b{}\b".format(word)
            if re.search(regexp, hocr, **search_kwargs):
                return True

        return False

    def _match_all(self, hocr, search_kwargs):

        for word in self._split_match():
            regexp = r"\b{}\b".format(word)
            search_result = re.search(
                regexp,
                hocr,
                **search_kwargs
            )
            if not search_result:
                return False

        return True

    def _match_literal(self, hocr, search_kwargs):
        """
        Simplest match - literal match  i.e.
        exact match of the given word or string.
        """
        regexp = r"\b{}\b".format(self.match)
        result = re.search(regexp, hocr, **search_kwargs)
        return bool(result)

    def _match_regexp(self, hocr, search_kwargs):
        regexp = re.compile(self.match, **search_kwargs)

        result = re.search(regexp, hocr)

        return bool(result)

    def _split_match(self):
        """
        Splits the match to individual keywords, getting rid of unnecessary
        spaces and grouping quoted words together.
        Example:
          '  some random  words "with   quotes  " and   spaces'
            ==>
          ["some", "random", "words", "with+quotes", "and", "spaces"]
        """
        findterms = re.compile(r'"([^"]+)"|(\S+)').findall
        normspace = re.compile(r"\s+").sub
        return [
            normspace(" ", (t[0] or t[1]).strip()).replace(" ", r"\s+")
            for t in findterms(self.match)
        ]



# Generated by Django 3.0.7 on 2020-07-21 08:25

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0015_auto_20200615_0525'),
    ]

    operations = [
        migrations.CreateModel(
            name='Automate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, unique=True)),
                ('match', models.CharField(blank=True, max_length=256)),
                ('matching_algorithm', models.PositiveIntegerField(choices=[(1, 'Any'), (2, 'All'), (3, 'Literal'), (4, 'Regular Expression')], default=1)),
                ('is_case_sensitive', models.BooleanField(default=True)),
                ('plugin_name', models.CharField(blank=True, max_length=256)),
                ('extract_page', models.BooleanField(default=False)),
                ('dst_folder', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='core.Folder')),
            ],
        ),
    ]



# Generated by Django 3.1.5 on 2021-01-18 06:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0036_auto_20210117_1449'),
    ]

    operations = [
        migrations.AddField(
            model_name='page',
            name='hocr_step_0',
            field=models.TextField(default=''),
        ),
        migrations.AddField(
            model_name='page',
            name='hocr_step_1',
            field=models.TextField(default=''),
        ),
        migrations.AddField(
            model_name='page',
            name='hocr_step_2',
            field=models.TextField(default=''),
        ),
        migrations.AddField(
            model_name='page',
            name='hocr_step_3',
            field=models.TextField(default=''),
        ),
    ]



import inspect

from django.apps import apps


class PartsFinder:
    """
    Finds models descending from one of:

        (1) papermerge.core.models.AbstractNode
        (2) papermerge.core.models.AbstractDocument
        (3) papermerge.core.models.AbstractFolder

    Looks for model candidates in all (current django) project apps.
    Models descending from (1), (2) ,(3) are called parts.
    """

    def find(self, abstract_klass):
        """
        Returns all models descendent from given ``abstract_klass``
        """
        app_configs = apps.get_app_configs()
        ret = []

        for app_config in app_configs:
            app_models = app_config.get_models()

            for model_klass in app_models:
                if self.descents_from_abstract(model_klass, abstract_klass):
                    ret.append(model_klass)

        return ret

    def get(self, abstract_klass, attr_name):
        """
        Returns model_klass which inherits from ``abstract_klass``
        AND has attribute ``attr_name``
        """
        all_parts = self.find(abstract_klass)

        for model_klass in all_parts:
            if hasattr(model_klass, attr_name):
                return model_klass

    def descents_from_abstract(self, klass, abstract_klass):
        """
        Is ``klass`` descending from ``abstract_klass``?
        """
        return abstract_klass in inspect.getmro(klass)


def get_parts_finder_class():
    return PartsFinder


parts_finder_class = get_parts_finder_class()

default_parts_finder = parts_finder_class()



import time
import logging
from django.conf import settings
from django.core.management.base import BaseCommand
from papermerge.core.importers.imap import import_attachment


logger = logging.getLogger(__name__)


class Command(BaseCommand):

    help = """Import documents from email attachments.
"""

    def add_arguments(self, parser):
        parser.add_argument(
            "--loop-time",
            "-t",
            default=settings.PAPERMERGE_IMPORTER_LOOP_TIME,
            type=int,
            help="Wait time between each loop (in seconds)."
        )

    def main_loop(self, loop_time):
        while True:
            start_time = time.time()

            import_attachment(
                imap_server=self._imap_server,
                username=self._username,
                password=self._password,
                delete=self._delete,
                inbox_name=self._inbox_name,
                by_user=self._by_user,
                by_secret=self._by_secret
            )

            # Sleep until the start of the next loop step
            time.sleep(
                max(0, start_time + loop_time - time.time())
            )

    def handle(self, *args, **options):
        loop_time = options.get('loop_time')
        # options used by importer
        self._imap_server = settings.PAPERMERGE_IMPORT_MAIL_HOST
        self._username = settings.PAPERMERGE_IMPORT_MAIL_USER
        self._password = settings.PAPERMERGE_IMPORT_MAIL_PASS
        self._delete = settings.PAPERMERGE_IMPORT_MAIL_DELETE
        self._by_user = settings.PAPERMERGE_IMPORT_MAIL_BY_USER
        self._by_secret = settings.PAPERMERGE_IMPORT_MAIL_BY_SECRET
        self._inbox_name = settings.PAPERMERGE_IMPORT_MAIL_INBOX

        try:
            self.main_loop(loop_time)
        except KeyboardInterrupt:
            logger.info("Exiting")



import logging

from django.contrib import messages
from django.urls import reverse
from django.contrib.auth.decorators import login_required
from django.shortcuts import render, redirect, get_object_or_404
from django.utils.translation import ugettext_lazy as _
from django.http import HttpResponseForbidden

from papermerge.core.models import User
from papermerge.contrib.admin.forms import (
    UserFormWithoutPassword,
    UserFormWithPassword
)

logger = logging.getLogger(__name__)


@login_required
def user_view(request):
    """
    When adding a new user, administrator will need to add
    username + password and then he/she will be able to edit further
    details.
    """

    if not request.user.is_superuser:
        return HttpResponseForbidden()

    form = UserFormWithPassword()

    if request.method == 'POST':
        form = UserFormWithPassword(request.POST)
        if form.is_valid():

            password1 = form.cleaned_data['password1']
            password2 = form.cleaned_data['password2']

            if password1 == password2:

                user = form.save()
                user.set_password(password1)
                user.save()

                return redirect(
                    reverse('core:user_change', args=(user.id, ))
                )
            else:
                form.add_error(
                    'password1',
                    _('Password and Password confirmation does not match')
                )

    return render(
        request,
        'admin/add_user.html',
        {
            'form': form,
        }
    )


@login_required
def user_change_view(request, id):
    """
    Used to edit existing users
    """

    if not request.user.is_superuser:
        return HttpResponseForbidden()

    user = get_object_or_404(User, id=id)
    action_url = reverse('core:user_change', args=(id,))

    form = UserFormWithoutPassword(
        request.POST or None, instance=user
    )

    if form.is_valid():
        form.save()
        return redirect('core:users')

    return render(
        request,
        'admin/user.html',
        {
            'form': form,
            'action_url': action_url,
            'title': _('Edit User'),
            'user_id': id
        }
    )


@login_required
def user_change_password_view(request, id):
    """
    This view is used by administrator to change password of ANY user in the
    system. As result, 'current password' won't be asked.
    """

    if not request.user.is_superuser:
        return HttpResponseForbidden()

    user = get_object_or_404(User, id=id)
    action_url = reverse('core:user_change_password', args=(id,))

    if request.method == 'POST':
        password1 = request.POST['password1']
        password2 = request.POST['password2']
        if password1 == password2:
            user.set_password(password1)
            user.save()
            messages.success(
                request,
                _("Password was successfully changed.")
            )
            return redirect(
                reverse('core:user_change', args=(id,))
            )

    return render(
        request,
        'admin/user_change_password.html',
        {
            'user': user,
            'action_url': action_url
        }
    )
